import torch
import torch.nn as nn
import numpy as np
import itertools
from torch.nn import CrossEntropyLoss
import math
from scipy import optimize as sopt
from torch.nn import functional as F
from torch.nn.functional import binary_cross_entropy


GEOMETRIES = ('cvar', 'chi-square')
MIN_REL_DIFFERENCE = 1e-5


def chi_square_value(p, v, reg):
    """Returns <p, v> - reg * chi^2(p, uniform) for Torch tensors"""
    m = p.shape[0]

    with torch.no_grad():
        chi2 = (0.5 / m) * reg * (
                torch.norm(m * p - torch.ones(m, device=v.device), p=2) ** 2)

    return torch.dot(p, v) - chi2


def cvar_value(p, v, reg):
    """Returns <p, v> - reg * KL(p, uniform) for Torch tensors"""
    m = p.shape[0]

    with torch.no_grad():
        idx = torch.nonzero(p)  # where is annoyingly backwards incompatible
        kl = np.log(m) + (p[idx] * torch.log(p[idx])).sum()

    return torch.dot(p, v) - reg * kl


def fenchel_kl_cvar(v, alpha):
    """Returns the empirical mean of the Fenchel dual for KL CVaR"""
    v -= np.log(1 / alpha)
    v1 = v[torch.lt(v, 0)]
    v2 = v[torch.ge(v, 0)]
    w1 = torch.exp(v1) / alpha - 1
    w2 = (v2 + 1) * (1 / alpha) - 1
    return (w1.sum() + w2.sum()) / v.shape[0]


def bisection(eta_min, eta_max, f, tol=1e-2, max_iter=500):
    """Expects f an increasing function and return eta in [eta_min, eta_max] 
    s.t. |f(eta)| <= tol (or the best solution after max_iter iterations"""
    lower = f(eta_min)
    upper = f(eta_max)

    # until the root is between eta_min and eta_max, double the length of the 
    # interval starting at either endpoint.
    while lower > 0 or upper < 0:
        length = eta_max - eta_min
        if lower > 0:
            eta_max = eta_min
            eta_min = eta_min - 2 * length
        if upper < 0:
            eta_min = eta_max
            eta_max = eta_max + 2 * length

        lower = f(eta_min)
        upper = f(eta_max)

    for _ in range(max_iter):
        eta = 0.5 * (eta_min + eta_max)

        v = f(eta)

        if torch.abs(v) <= tol:
            return eta

        if v > 0:
            eta_max = eta
        elif v < 0:
            eta_min = eta
    return 0.5 * (eta_min + eta_max)


def huber_loss(x, delta=1.):
    """ Standard Huber loss of parameter delta

    https://en.wikipedia.org/wiki/Huber_loss

    returns 0.5 * x^2 if |a| <= \delta
            \delta * (|a| - 0.5 * \delta) o.w.
    """
    if torch.abs(x) <= delta:
        return 0.5 * (x ** 2)
    else:
        return delta * (torch.abs(x) - 0.5 * delta)


class RobustLoss(nn.Module):
    """PyTorch module for the batch robust loss estimator"""
    def __init__(self, size, reg, geometry, tol=1e-2,
                 max_iter=500, debugging=False, is_regression=False, device=torch.device("cuda:6")):
        """
        Parameters
        ----------

        size : float
            Size of the uncertainty set (\rho for \chi^2 and \alpha for CVaR)
            Set float('inf') for unconstrained
        reg : float
            Strength of the regularizer, entropy if geometry == 'cvar'
            $\chi^2$ divergence if geometry == 'chi-square'
        geometry : string
            Element of GEOMETRIES
        tol : float, optional
            Tolerance parameter for the bisection
        max_iter : int, optional
            Number of iterations after which to break the bisection

        """
        super().__init__()
        self.size = size
        self.reg = reg
        self.geometry = geometry
        self.tol = tol
        self.max_iter = max_iter
        self.debugging = debugging
        self.is_regression = is_regression
        self.is_erm = size == 0
        self.device = device 
        

        if geometry not in GEOMETRIES:
            raise ValueError('Geometry %s not supported' % geometry)

        if geometry == 'cvar' and self.size > 1:
            raise ValueError(f'alpha should be < 1 for cvar, is {self.size}')

    def best_response(self, v):
        size = self.size
        reg = self.reg
        m = v.shape[0]

        if self.geometry == 'cvar':
            if self.reg > 0:
                if size == 1.0:
                    return torch.ones_like(v) / m

                def p(eta):
                    x = (v - eta) / reg
                    return torch.min(torch.exp(x),
                                     torch.Tensor([1 / size]).type(x.dtype).to(self.device)) / m

                def bisection_target(eta):
                    return 1.0 - p(eta).sum()

                eta_min = reg * torch.logsumexp(v / reg - np.log(m), 0)
                eta_max = v.max()

                if torch.abs(bisection_target(eta_min)) <= self.tol:
                    return p(eta_min)
            else:
                cutoff = int(size * m)
                surplus = 1.0 - cutoff / (size * m)

                p = torch.zeros_like(v).to(self.device)
                idx = torch.argsort(v, descending=True)
                p[idx[:cutoff]] = 1.0 / (size * m)
                if cutoff < m:
                    p[idx[cutoff]] = surplus
                return p

        if self.geometry == 'chi-square':
            if (v.max() - v.min()) / v.max() <= MIN_REL_DIFFERENCE:
                return (torch.ones_like(v) / m).to(self.device)

            if size == float('inf'):
                assert reg > 0

                def p(eta):
                    return (torch.relu(v - eta) / (reg * m)).to(device)

                def bisection_target(eta):
                    return 1.0 - p(eta).sum()

                eta_min = min(v.sum() - reg * m, v.min())
                eta_max = v.max()

            else:
                assert size < float('inf')

                # failsafe for batch sizes small compared to
                # uncertainty set size
                if m <= 1 + 2 * size:
                    out = (v == v.max()).float()
                    out /= out.sum()
                    return out

                if reg == 0:
                    def p(eta):
                        pp = torch.relu(v - eta)
                        return pp / pp.sum()

                    def bisection_target(eta):
                        pp = p(eta)
                        w = m * pp - torch.ones_like(pp)
                        return 0.5 * torch.mean(w ** 2) - size

                    eta_min = -(1.0 / (np.sqrt(2 * size + 1) - 1)) * v.max()
                    eta_max = v.max()
                else:
                    def p(eta):
                        pp = torch.relu(v - eta)

                        opt_lam = max(
                            reg, torch.norm(pp) / np.sqrt(m * (1 + 2 * size))
                        )

                        return pp / (m * opt_lam)

                    def bisection_target(eta):
                        return 1 - p(eta).sum()

                    eta_min = v.min() - 1
                    eta_max = v.max()

        eta_star = bisection(
            eta_min, eta_max, bisection_target,
            tol=self.tol, max_iter=self.max_iter)

        if self.debugging:
            return p(eta_star), eta_star
        return p(eta_star)

    def forward(self, outputs, targets):
        """Value of the robust loss

        Note that the best response is computed without gradients

        Parameters
        ----------

        v : torch.Tensor
            Tensor containing the individual losses on the batch of examples

        Returns
        -------
        loss : torch.float
            Value of the robust loss on the batch of examples
        """

        if self.is_regression:
            v = (outputs.squeeze() - targets) ** 2
        else:
            losstemp = CrossEntropyLoss(reduction="none")
            v = losstemp(outputs, targets.long())

        if self.is_erm:
            return v.mean()
        else:
            with torch.no_grad():
                p = self.best_response(v)

            if self.geometry == 'cvar':
                return cvar_value(p, v, self.reg)
            elif self.geometry == 'chi-square':
                return chi_square_value(p, v, self.reg)

def chi_square_doro_criterion(outputs, targets, alpha, eps):
    batch_size = len(targets)
    criterion = CrossEntropyLoss(reduction="none")
    loss = criterion(outputs, targets, )
    # Chi^2-DORO
    max_l = max(10., int(0.6*loss.max().item()))
    C = math.sqrt(1 + (1 / alpha - 1) ** 2)
    n = int(eps * batch_size)
    rk = torch.argsort(loss, descending=True)
    l0 = loss[rk[n:]]
    foo = lambda eta: C * math.sqrt(
        (F.relu(l0.cpu() - eta) ** 2).mean().item()) + eta
    # print("begin opt search")
    # opt_eta = sopt.brent(foo, brack=(0, max_l),tol=1e-2)
    opt_eta = sopt.brent(foo, brack=(0, max_l))
    
    # print("end opt search")
    loss = C * torch.sqrt((F.relu(l0 - opt_eta) ** 2).mean()) + opt_eta
    return loss


# DORO; adapted from
# https://github.com/RuntianZ/doro/blob/master/wilds-exp/algorithms/doro.py
def cvar_doro_criterion(outputs, targets, eps, alpha):
    batch_size = len(targets)
    criterion = CrossEntropyLoss(reduction="none")
    loss = criterion(outputs, targets)
    # CVaR-DORO
    gamma = eps + alpha * (1 - eps)
    n1 = int(gamma * batch_size)
    n2 = int(eps * batch_size)
    rk = torch.argsort(loss, descending=True)
    loss = loss[rk[n2:n1]].sum() / alpha / (batch_size - n2)
    return loss

def get_dro_criterion(size, reg, geometry, tol=1e-4,max_iter=1000, is_regression=False):
    return RobustLoss(
            geometry=geometry,
            size=size,
            reg=reg,
            max_iter=max_iter,
            is_regression=is_regression
        )
    

def group_dro_criterion(outputs, targets, sens):
    # TODO: adjust these setups to not binary group??
    assert torch.all(torch.logical_or(sens == 1., sens == 2.)), \
        "only binary groups supported."
    subgroup_losses = []
    n_attrs = sens.shape[1]
    criterion = CrossEntropyLoss(reduction="none")
    elementwise_loss = criterion(outputs, targets)
    # Compute the loss on each subgroup
    for subgroup_idxs in itertools.product(*[(1, 2)] * n_attrs):
        subgroup_idxs = torch.Tensor(subgroup_idxs).to(sens.device)
        mask = torch.all(sens == subgroup_idxs, dim=1)
        subgroup_loss = elementwise_loss[mask].sum() / mask.sum()
        subgroup_losses.append(subgroup_loss)
    return torch.stack(subgroup_losses)




from dataclasses import dataclass
import pathlib


@dataclass
class FilesResource:
    """Stores information needed to (down)load a dataset.

    Attributes
    ----------
    url : str
        The URL from where the dataset will be downloaded.
    download_path : pathlib.Path
        Where we'll be saving the content of the HTTP request.
    file_name : str
        Name of the file where the dataset is stored in local memory.
    data_dir : str
        Name of the directory where the dataset is stored in local memory.
    """
    url: str
    download_path: pathlib.Path
    file_name: str
    data_dir: str

    @property
    def file_path(self):
        """Generates a pathlib.Path object pointing to where the data
        is (will be) stored based on the `file_name` and `data_dir`
        properties.

        Returns
        -------
        pathlib.Path
            The path to where the file is (will be) stored.
        """
        return pathlib.Path(self.data_dir, self.file_name)



from scipy.optimize import brent
import torch
import torch.nn as nn
import numpy as np
from tqdm import tqdm
from sklearn.metrics import f1_score
from torch.autograd import grad
from sklearn.ensemble import GradientBoostingClassifier, RandomForestClassifier, RandomForestRegressor
from sklearn import linear_model, ensemble, kernel_approximation, svm
from sklearn.utils import resample
from sklearn.metrics import accuracy_score
import xgboost as xgb
from fairlearn.reductions import DemographicParity, EqualizedOdds, \
    ErrorRateParity
from fairlearn.postprocessing import ThresholdOptimizer
from fairlearn.reductions import ExponentiatedGradient, DemographicParity
from torch.nn import CrossEntropyLoss
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torch.utils.data import random_split
from .robust_loss import RobustLoss, group_dro_criterion, cvar_doro_criterion, chi_square_doro_criterion
from .marginal_dro_criterion import LipLoss, opt_model, marginal_dro_criterion
from .model_util import *


TRAIN_FRAC = 0.8

class MLP(nn.Module):
    def __init__(self, input_dim, num_classes, num_units=16, nonlin=nn.ReLU(), dropout_ratio=0.1):
        super().__init__()

        self.dense0 = nn.Linear(input_dim, num_units)
        self.nonlin = nonlin
        self.dropout = nn.Dropout(dropout_ratio)
        self.dense1 = nn.Linear(num_units, num_units)
        self.output = nn.Linear(num_units, num_classes)

    def forward(self, X, **kwargs):
        X = self.nonlin(self.dense0(X.float()))
        X = self.dropout(X)
        X = self.nonlin(self.dense1(X))
        X = self.output(X)
        return X

class Linear(nn.Module):
    def __init__(self, input_dim, num_classes=2):
        super().__init__()
        self.dense0 = nn.Linear(input_dim, num_classes)
        
    def forward(self, X, **kwargs):
        return self.dense0(X.float())

class LogisticRegression():
    def __init__(self, input_dim=9, num_classes=2):
        self.model = Linear(input_dim, num_classes)
        self.criterion = nn.CrossEntropyLoss()

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
    
    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total
    
    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='micro')


    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        # criterion = nn.CrossEntropyLoss()
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=8)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=8)
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        val_loss = 0.0
        val_steps = 0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

                loss = self.criterion(outputs, labels)
                val_loss += loss.cpu().numpy()
                val_steps += 1

        # print("accuracy", correct / total)

        return correct / total

class MLPClassifier():
    def __init__(self, input_dim=9, num_classes=2):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.model = MLP(input_dim, num_classes)
        self.criterion = nn.CrossEntropyLoss()

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
    
    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def predict_proba(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        # print(outputs.shape)
        return outputs.detach().cpu().numpy()


    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        # criterion = nn.CrossEntropyLoss()
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        for epoch in tqdm(range(0,self.train_epochs+1)):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

    def fit_weight(self, X, y, weights, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 
        self.criterion = nn.CrossEntropyLoss(reduction='none')
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
            weights = torch.tensor(weights).float()

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        # criterion = nn.CrossEntropyLoss()
        
        trainset = TensorDataset(X, y, weights)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        for epoch in tqdm(range(0,self.train_epochs+1)):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels, weights_batch = data
                inputs, labels, weights_batch = inputs.to(self.device), labels.to(self.device), weights_batch.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                loss = torch.dot(loss.reshape(-1), weights_batch.reshape(-1))/loss.shape[0]
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels, weights_batch = data
                inputs, labels, weights_batch = inputs.to(self.device), labels.to(self.device), weights_batch.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

#chi-DRO
class chi_DRO():
    def __init__(self, input_dim=9, num_classes=2, size=0.1, reg=0.1, max_iter=100, device='cpu'):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.model = MLP(input_dim, num_classes)
        self.criterion = RobustLoss(
            geometry='chi-square',
            size=size,
            reg=reg,
            max_iter=max_iter,
            is_regression=False, device=torch.device("cuda:6"))
        self.device = device

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.criterion = RobustLoss(
            geometry='chi-square',
            size=config["size"],
            reg=config["reg"],
            max_iter=500,
            is_regression=False, device=config["device"])

    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        # criterion = nn.CrossEntropyLoss()
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True)
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

#CVaR-DRO
class cvar_DRO():
    def __init__(self, input_dim=9, num_classes=2, size=0.1, reg=0.1, max_iter=100):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.model = MLP(input_dim, num_classes)
        self.criterion = RobustLoss(
            geometry='cvar',
            size=size,
            reg=reg,
            max_iter=max_iter,
            is_regression=False, device=torch.device("cuda:6"))

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.device = config["device"]
        self.criterion = RobustLoss(
            geometry='cvar',
            size=config["size"],
            reg=config["reg"],
            max_iter=config["max_iter"],
            is_regression=False, device=config["device"])

    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device=torch.device("cpu")):
        assert self.device == device
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        # criterion = nn.CrossEntropyLoss()
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        # print("Begin fitting..")
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = self.criterion(outputs, labels)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                # print(i)
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

#CVaR-DORO
class cvar_DORO():
    def __init__(self, input_dim=9, num_classes=2):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.model = MLP(input_dim, num_classes)

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.eps = config["eps"]
        self.alpha = config["alpha"]


    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = cvar_doro_criterion(outputs, labels, self.eps, self.alpha)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        val_loss = 0.0
        val_steps = 0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

#chi-DORO
class chi_DORO():
    def __init__(self, input_dim=9, num_classes=2, eps=0.1, alpha=0.1,lr=1e-3):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.model = MLP(input_dim, num_classes)
        self.eps = eps 
        self.alpha = alpha 
        self.lr = lr
        self.batch_size=128
        self.train_epochs=100

    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.eps = config["eps"]
        self.alpha = config["alpha"]


    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        for epoch in range(0,self.train_epochs+1):
            # print(epoch)
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = chi_square_doro_criterion(outputs, labels, self.eps, self.alpha)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        val_loss = 0.0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

#GROUP-DRO
class group_DRO():
    def __init__(self, input_dim=9, num_classes=2, n_groups = 2, device='cpu'):
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.sensitive_feature = [0]
        self.model = MLP(input_dim, num_classes)
        self.device = device
        self.group_weights = torch.ones(n_groups, device=self.device)
        self.group_weights = self.group_weights / self.group_weights.sum()



    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.sensitive_feature = config["sensitive_features"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.group_weights_step_size = torch.Tensor(
            [config["group_weights_step_size"]]).to(self.device)



    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total
    
    def f1score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return f1_score(y.cpu().numpy(), predicted.cpu().numpy(), average='macro')

    def group_info(self, X):
        if isinstance(X, torch.Tensor):
            return torch.Tensor(X.cpu().numpy()[:,self.sensitive_feature]).to(self.device)
        else:
            raise NotImplementedError

    def fit(self, X, y, train_ratio=TRAIN_FRAC, device = "cpu"):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=1)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=1)
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                g = self.group_info(inputs)
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                optimizer.zero_grad()
                outputs = self.model(inputs)
                group_losses = group_dro_criterion(outputs, labels, g)
                # update group weights
                self.group_weights = self.group_weights * torch.exp(
                    self.group_weights_step_size * group_losses.data)
                self.group_weights = (self.group_weights / (self.group_weights.sum()))
                # update model
                loss = group_losses @ self.group_weights
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        val_loss = 0.0
        val_steps = 0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)

                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

# marginal-dro
class marginal_DRO():
    def __init__(self, input_dim=9, num_classes=2, device='cpu'):
        self.input_dim = input_dim
        self.num_classes = num_classes
        # set this to be the shifted covariate
        self.sensitive_feature = [0]
        self.model = MLP(input_dim, num_classes)
        self.device = device
    
    def update(self, config):
        self.lr = config["lr"]
        self.batch_size = config["batch_size"]
        self.train_epochs = config["train_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.radius = config["radius"]
        self.p_min = config["p_min"]
        self.nbisect = config["nbisect"]
        self.max_iter = config["max_iter"]
    
    @property
    def rho(self):
        return 1.0 / self.p_min

    def predict(self, X):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
        inputs = X.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        return predicted.detach().cpu().numpy()

    def score(self, X, y):
        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)
        inputs, labels = X.to(self.device), y.to(self.device)
        self.model = self.model.to(self.device)
        outputs = self.model(inputs)
        _, predicted = torch.max(outputs.data, 1)
        correct = (predicted == labels).sum().item()
        total = y.shape[0]
        return correct / total

    def group_info(self, X):
        if isinstance(X, torch.Tensor):
            return torch.Tensor(X.cpu().numpy()[:,self.sensitive_feature]).to(self.device)
        else:
            raise NotImplementedError
    def _init_eta(self, x_in, y_in):
        loss = nn.CrossEntropyLoss()
        sens = self.group_info(x_in)
        criterion = LipLoss(radius=self.radius, x_in=sens,
                       eta=0)
        wrapped_fun = lambda eta: opt_model(self.model, loss, criterion, 0.0,
                                            self.rho, x_in=x_in, y_in=y_in,
                                            lr=self.lr,
                                            niter=self.max_iter)[0][-1]
        
        opt_init = opt_model(self.model, loss, criterion, 0.0, self.rho, x_in,
                             y_in, lr=self.lr, niter=self.max_iter)
        brack_ivt = (min(0, np.nanmin(opt_init[1])), np.nanmax(opt_init[1]))
        bopt = brent(wrapped_fun, brack=brack_ivt, maxiter=self.nbisect,
                     full_output=True)
        eta = bopt[0]
        return eta


    def fit(self, X, y, train_ratio=TRAIN_FRAC, device='cpu'):
        self.device = device 

        if not isinstance(X, torch.Tensor):
            X = torch.tensor(X)
            y = torch.tensor(y)

        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        
        trainset = TensorDataset(X, y)
        test_abs = int(len(trainset) * train_ratio)
        train_subset, val_subset = random_split(trainset, [test_abs, len(trainset) - test_abs])
        trainloader = torch.utils.data.DataLoader(train_subset,batch_size=self.batch_size, shuffle=True, num_workers=8)
        valloader = torch.utils.data.DataLoader(val_subset, batch_size=self.batch_size, shuffle=True, num_workers=8)
        for epoch in range(0,self.train_epochs+1):
            running_loss = 0.0
            epoch_steps = 0
            for i, data in enumerate(trainloader, 0):
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                g = self.group_info(inputs)
                optimizer.zero_grad()
                outputs = self.model(inputs)
                self.eta_train = self._init_eta(inputs, labels)
                loss = marginal_dro_criterion(outputs, labels, g, self.radius, self.eta_train)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()
                epoch_steps += 1
                
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        val_loss = 0.0
        val_steps = 0
        total = 0
        correct = 0
        for i, data in enumerate(valloader, 0):
            with torch.no_grad():
                inputs, labels = data
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                
                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        # print("accuracy", correct / total)

        return correct / total

# SVM WITH KERNEL APPROXIMATION
class SVM2():
    def __init__(self):
        self.model = svm.LinearSVC()
        self.preprocessor = kernel_approximation.RBFSampler()
    def update(self, config):
        self.kernel = config['kernel']
        
        if config["kernel"] == 'rbf':
            self.preprocessor = kernel_approximation.RBFSampler(gamma = config["gamma"], n_components = config["n_components"], random_state = 1)
            print(self.preprocessor)
        elif config["kernel"] == 'poly':
            self.preprocessor == kernel_approximation.PolynomialCountSketch(gamma = config["gamma"], n_components = config["n_components"])
        elif config["kernel"] == 'linear':
            pass
        self.model = svm.LinearSVC()
    def fit(self, X, y, device=None):
        if self.kernel != 'linear':
            X = self.preprocessor.fit_transform(X)
        self.model.fit(X, y)

    def predict(self, X):
        if self.kernel != 'linear':
            X = self.preprocessor.fit_transform(X)
        return self.model.predict(X)
    def score(self, X, y):
        predicted = self.predict(X)
        correct = (predicted == y).sum()
        total = y.shape[0]
        return correct / total

# JTT approaches (reuse models), for xgboost models we directly change the sample weight for each learner
class JTT():
    def __init__(self, sensitive_features=[0], base_method='xgb'):
        self.sensitive_features = sensitive_features
        if base_method == 'xgb':
            base_model = xgb.XGBClassifier()
        elif base_method == 'gbm':
            base_model = GradientBoostingClassifier()
        else:
            raise NotImplementedError
        self.model = base_model
    def update(self, config):
        self.model = xgb.XGBClassifier(learning_rate=config["learning_rate"], min_split_loss=config["min_split_loss"],
                    max_depth=config["max_depth"], colsample_bytree=config["colsample_bytree"], colsample_bylevel=config["colsample_bylevel"],
                    max_bin=config["max_bin"], grow_policy=config["grow_policy"])
        self.up_weight = config['lambda']



    def fit(self, X, y, train_ratio=TRAIN_FRAC, device = None):
        train_num = int(train_ratio*X.shape[0])
        trainX = X[:train_num,:]
        trainy = y[:train_num]
        valX = X[train_num:, :]
        valy = y[train_num:]
        # first stage
        self.model.fit(trainX, trainy)
        predicted =  self.model.predict(trainX)
        sample_weight = [self.up_weight if predicted[i] != y[i] else 1 for i in range(len(trainX))]
        # second stage: upweight + retrain
        self.model.fit(trainX, trainy, sample_weight = sample_weight)

        return self.score(valX, valy)
    def predict(self, X):
        return self.model.predict(X)

    def score(self, X, y):
        predicted = self.predict(X)
        correct = (predicted == y).sum()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        predicted = self.predict(X)
        return f1_score(y, predicted, average='macro')

# subsampling approaches
class SubSample():
    def __init__(self, X, y, sensitive_features = [0]):
    # transform to pd.DataFrame
        self.X = X
        self.y = y
        self.sensitive_feature = sensitive_features[0]
        self.df_data = pd.DataFrame(X)
        self.df_data['label'] = y
        pass
    def reload(self, balance_name):
        if balance_name == 'suby':
            if np.mean(self.y) > 1:
                grp_pos = resample(self.df_data[self.df_data['label'] == 1], replace = True, n_samples = len(self.df_data[self.df_data['label'] == 0]))
                new_data = pd.concat([grp_pos, self.df_data[self.df_data['label'] == 0]])
            else:
                grp_pos = resample(self.df_data[self.df_data['label'] == 0], replace = True, n_samples = len(self.df_data[self.df_data['label'] == 1]))
                new_data = pd.concat([grp_pos, self.df_data[self.df_data['label'] == 1]])
            y = new_data['label'].to_numpy()
            del new_data['label']
            X = np.array(new_data)
            sample_weight = np.ones(len(X))

        elif balance_name == 'subg':

            feature_cnt = self.df_data.groupby(self.sensitive_feature).count()['label']
            min_cnt, min_cnt_grp = np.min(feature_cnt), np.argmin(feature_cnt)

            new_data = self.df_data[self.df_data[self.sensitive_feature] == feature_cnt.index[min_cnt_grp]]
            for i, idx in enumerate(feature_cnt.index):
                if i != min_cnt_grp:
                    grp_data = resample(self.df_data[self.df_data[self.sensitive_feature] == idx], replace = True, n_samples = min_cnt)
                    new_data = pd.concat([new_data, grp_data])    
            y = new_data['label'].to_numpy()
            del new_data['label']
            X = np.array(new_data)
            sample_weight = np.ones(len(X))

        
        elif balance_name == 'rwy':
            X = self.X
            y = self.y
            up_weight = np.mean(y) / (1 - np.mean(y))
            sample_weight = [1 if y[i] ==  1 else up_weight for i in range(len(y))]
            sample_weight = (sample_weight / np.sum(sample_weight))*X.shape[0]

        elif balance_name == 'rwg':
            X = self.X
            y = self.y
            feature_cnt = self.df_data.groupby(self.sensitive_feature).count()['label']
            min_cnt, min_cnt_grp = np.min(feature_cnt), np.argmin(feature_cnt)
            sample_weightdict = dict.fromkeys(feature_cnt.index)
            for i, idx in enumerate(feature_cnt.index):
                sample_weightdict[idx] = min_cnt / list(feature_cnt)[i]
            sample_weight = [sample_weightdict[i] for i in self.df_data[self.sensitive_feature]]
            sample_weight = (sample_weight / np.sum(sample_weight))*X.shape[0]

        else:
            raise NotImplementedError
        return X, y, sample_weight

# fairness approaches
class FairPostprocess():
    def __init__(self, sensitive_features=[0], base_method='xgb'):
        self.sensitive_features = sensitive_features
        if base_method == 'xgb':
            base_model = xgb.XGBClassifier()
        elif base_method == 'gbm':
            base_model = GradientBoostingClassifier()
        else:
            raise NotImplementedError
        self.model = ThresholdOptimizer(estimator=base_model, predict_method='predict')

    def update(self, config):
        base_model = xgb.XGBClassifier(learning_rate=config["learning_rate"], min_split_loss=config["min_split_loss"],
                    max_depth=config["max_depth"], colsample_bytree=config["colsample_bytree"], colsample_bylevel=config["colsample_bylevel"],
                    max_bin=config["max_bin"], grow_policy=config["grow_policy"])
        self.kind = config["kind"]
        self.sensitive_features = config["sensitive_features"]
        if config["kind"] == "threshold":
            self.model = ThresholdOptimizer(estimator=base_model, predict_method='predict')
        elif config["kind"] == 'exp':
            self.model = ExponentiatedGradient(estimator=base_model, constraints=DemographicParity(difference_bound=0.02))
        else:
            raise NotImplementedError
    
    def fit(self, X, y, train_ratio=TRAIN_FRAC, device=None):
        train_num = int(train_ratio*X.shape[0])
        trainX = X[:train_num,:]
        trainy = y[:train_num]
        valX = X[train_num:, :]
        valy = y[train_num:]
        self.model.fit(trainX, trainy, sensitive_features=trainX[:,self.sensitive_features])
        return self.score(valX, valy)
    
    def predict(self, X):
        if self.kind == "threshold":
            return self.model.predict(X, sensitive_features=X[:,self.sensitive_features],
                               random_state=0)
        elif self.kind == "exp":
            return self.model.predict(X, random_state=0)
    
    def score(self, X, y):
        predicted = self.predict(X)
        correct = (predicted == y).sum()
        total = y.shape[0]
        return correct / total

    def f1score(self, X, y):
        predicted = self.predict(X)
        return f1_score(y, predicted, average='macro')



class FairInprocess():
    def __init__(self, sensitive_features = [0], base_method = 'xgb'):
        self.sensitive_features = sensitive_features
        if base_method == 'xgb':
            base_model = xgb.XGBClassifier()
        elif base_method == 'gbm':
            base_model = GradientBoostingClassifier()
        else:
            raise NotImplementedError
        self.model = CustomExponentiatedGradient(estimator=base_model,
                                    constraints = EqualizedOdds(),
                                    sensitive_features=self.sensitive_features)
        
        
    def update(self, config):
        base_model = xgb.XGBClassifier(learning_rate=config["learning_rate"], min_split_loss=config["min_split_loss"],
                    max_depth=config["max_depth"], colsample_bytree=config["colsample_bytree"], colsample_bylevel=config["colsample_bylevel"],
                    max_bin=config["max_bin"], grow_policy=config["grow_policy"])
        self.kind = config["kind"]
        self.sensitive_features = config["sensitive_features"]
        # Constraint
        if config["kind"] == "dp":
            constraint = DemographicParity()
        elif config["kind"] == "eo":
            constraint = EqualizedOdds()
        elif config["kind"] == "error_parity":
            constraint = ErrorRateParity()
        else:
            raise NotImplementedError
        # Model
        self.model = CustomExponentiatedGradient(estimator=base_model,
                                            constraints=constraint,
                                            sensitive_features=self.sensitive_features)
    def fit(self, X, y, train_ratio=TRAIN_FRAC, device=None):
        train_num = int(train_ratio*X.shape[0])
        trainX = X[:train_num,:]
        trainy = y[:train_num]
        valX = X[train_num:, :]
        valy = y[train_num:]
        self.model.fit(trainX, trainy)
        return self.score(valX, valy)
    
    def predict(self, X):
        z = self.model.predict_proba(X)
        return np.argmax(z, axis=1)


    def score(self, X, y):
        predicted = self.predict(X)
        correct = (predicted == y).sum()
        total = y.shape[0]
        return correct / total
    
    def f1score(self, X, y):
        predicted = self.predict(X)
        return f1_score(y, predicted, average='macro')

class IRM():
    def __init__(self, input_dim=9, num_classes=2, alpha=0.1, pretrain_epochs=100):
        self.model = MLP(input_dim, num_classes)
        self.alpha=alpha
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.pretrain_epochs = pretrain_epochs
        self.lr = 1e-3
        self.device = torch.device("cuda:7")
        self.train_epochs=100
    
    def update(self, config):
        self.lr = config["lr"]
        self.train_epochs = config["train_epochs"]
        self.pretrain_epochs = config["pretrain_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.alpha = config["alpha"]
    
    def fit(self, X_list, y_list, train_ratio=TRAIN_FRAC, device=torch.device("cuda:7")):
        self.device = device 

        if not isinstance(X_list[0], torch.Tensor):
            X_list = [torch.tensor(x) for x in X_list]
            y_list = [torch.tensor(y) for y in y_list]

        X_train_list = [x[:int(x.shape[0]*train_ratio),:] for x in X_list]
        X_val_list = [x[int(x.shape[0]*train_ratio):,:] for x in X_list]
        y_train_list = [y[:int(y.shape[0]*train_ratio)] for y in y_list]
        y_val_list = [y[int(y.shape[0]*train_ratio):] for y in y_list]
        
        self.model = self.model.to(self.device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        criterion = nn.CrossEntropyLoss()
        
        for epoch in range(0,self.train_epochs+1):
            scale = torch.tensor(1.).to(self.device).requires_grad_()
            penalty = 0.0
            loss = 0.0
            for idx in range(len(X_train_list)):
                inputs, labels = X_train_list[idx].to(self.device), y_train_list[idx].to(self.device)

                yhat = self.model(inputs)
                loss += criterion(yhat, labels)
                penalty += grad(criterion(yhat * scale, labels), [scale], create_graph=True)[0].pow(2).mean()
            
            irm_lam = self.alpha if epoch > self.pretrain_epochs else 1.0
            
            loss += irm_lam * penalty
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        accs = 0.0
        for idx in range(len(X_val_list)):
            with torch.no_grad():
                inputs, labels = X_val_list[idx].to(self.device), y_val_list[idx].to(self.device)
                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                accs += (predicted == labels).sum().item()/inputs.shape[0]

        # print("accuracy", accs / len(X_val_list))

        return accs / len(X_val_list)

class IGA():
    def __init__(self, input_dim=9, num_classes=2, alpha=0.1, pretrain_epochs=100):
        self.model = MLP(input_dim, num_classes)
        self.alpha=alpha
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.pretrain_epochs = pretrain_epochs
        self.lr = 1e-3
        self.device = torch.device("cuda:7")
        self.train_epochs=100
    
    def update(self, config):
        self.lr = config["lr"]
        self.train_epochs = config["train_epochs"]
        self.pretrain_epochs = config["pretrain_epochs"]
        self.model = MLP(self.input_dim, self.num_classes, num_units=config["hidden_size"], dropout_ratio=config["dropout_ratio"])
        self.alpha = config["alpha"]
    
    def fit(self, X_list, y_list, train_ratio=TRAIN_FRAC, device=torch.device("cuda:7")):
        self.device = device 

        if not isinstance(X_list[0], torch.Tensor):
            X_list = [torch.tensor(x) for x in X_list]
            y_list = [torch.tensor(y) for y in y_list]

        X_train_list = [x[:int(x.shape[0]*train_ratio),:] for x in X_list]
        X_val_list = [x[int(x.shape[0]*train_ratio):,:] for x in X_list]
        y_train_list = [y[:int(y.shape[0]*train_ratio)] for y in y_list]
        y_val_list = [y[int(y.shape[0]*train_ratio):] for y in y_list]
        
        self.model = self.model.to(device)
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        criterion = nn.CrossEntropyLoss()
        
        for epoch in range(0,self.train_epochs+1):
            scale = torch.tensor(1.).to(self.device).requires_grad_()
            penalty = 0.0
            loss = 0.0
            grad_list = []
            for idx in range(len(X_train_list)):
                inputs, labels = X_train_list[idx].to(device), y_train_list[idx].to(device)

                yhat = self.model(inputs)
                loss += criterion(yhat, labels)
                grad_list.append(grad(criterion(yhat * scale, labels), [scale], create_graph=True)[0])
            
            grads = torch.stack(grad_list)
            
            penalty = torch.var(grads, dim=0).sum()


            irm_lam = self.alpha if epoch > self.pretrain_epochs else 1.0
            
            loss += irm_lam * penalty
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            # print("[%d] loss: %.3f" % (epoch + 1,running_loss))
            
            
        accs = 0.0
        for idx in range(len(X_val_list)):
            with torch.no_grad():
                inputs, labels = X_val_list[idx].to(self.device), y_val_list[idx].to(self.device)
                outputs = self.model(inputs)
                _, predicted = torch.max(outputs.data, 1)
                accs += (predicted == labels).sum().item()/inputs.shape[0]

        # print("accuracy", accs / len(X_val_list))

        return accs / len(X_val_list)


class DWRPreprocess():
    def __init__(self):
        self.base_model = xgb.XGBClassifier()
        self.device = torch.device("cuda:7")
        self.lmbd_l1 = 0.1
        self.lmbd_l2 = 0.1
        self.weight = None
        self.threshold = 1e-3
        
    def update(self, config):
        self.base_model = xgb.XGBClassifier(learning_rate=config["learning_rate"], min_split_loss=config["min_split_loss"],
                    max_depth=config["max_depth"], colsample_bytree=config["colsample_bytree"], colsample_bylevel=config["colsample_bylevel"],
                    max_bin=config["max_bin"], grow_policy=config["grow_policy"])
        self.device = config["device"]
        self.weight = None
        self.lmbd_l1 = config["l1"]
        self.lmbd_l2 = config["l2"]
        self.threshold = config["threshold"]

    def covfunc(self, x, w=None):
        if w is None:
            n = x.shape[0]
            cov = torch.matmul(x.T, x) / n
            e = torch.mean(x, dim=0).view(-1, 1)
            res = cov - torch.matmul(e, e.T)
        else:
            w = w.view(-1, 1)
            cov = torch.matmul((w * x).T, x)
            e = torch.sum(w * x, dim=0).view(-1, 1)
            res = cov - torch.matmul(e, e.T)
        return res

    def corr(self, x, w=None):
        if w is None:
            n = x.shape[0]
            w = torch.ones(n) / n
        w = w.view(-1, 1)
        covariance = self.covfunc(x, w)
        var = torch.sum(w * (x * x), dim=0).view(-1, 1) - torch.square(
            torch.sum(w * x, dim=0).view(-1, 1)
        )
        index = torch.where(var<=0)[0]
        # if torch.min(var) < 0:
        #     print("error: ", torch.min(var))
        std = torch.sqrt(var)
        res = covariance
        # res = covariance / torch.matmul(std, std.T)
        return res

    def decorrelate_loss(self, x, weight, mask):
        assert weight.shape == (x.shape[0], 1)

        w = weight * weight
        n, p = x.shape
        res = self.corr(x, w / torch.sum(w))
        res = res * mask
        loss = torch.mean(res * res)
        
        balance_loss = loss
        loss_weight_sum = (torch.mean(w) - 1) ** 2
        loss_weight_l2 = torch.mean(w ** 2)
        loss = (
            1000 * loss + self.lmbd_l1 * loss_weight_sum + self.lmbd_l2 * loss_weight_l2
        )
        return loss, balance_loss, loss_weight_sum, loss_weight_l2

    def reweight(self, X):
        if self.weight is None:
            x = torch.from_numpy(X).type(torch.float).to(self.device)

            self.weight = torch.ones(x.shape[0], 1).to(self.device)
            num_step = 2000
            tot = 1e-8
            self.weight.requires_grad = True
            mask = 1 - torch.eye(x.shape[1]).to(self.device)

            optimizer = torch.optim.Adam([self.weight], lr=0.001)
            l_pre = 0
            for i in range(num_step):
                optimizer.zero_grad()
                loss, balance_loss, loss_s, loss_2 = self.decorrelate_loss(
                    x, self.weight, mask
                )
                loss.backward()
                optimizer.step()
                if torch.abs(loss - l_pre) <= tot:
                    break
                l_pre = loss


            self.weight.requires_grad = False
            # print('mean', self.weight.mean())
            self.weight = self.weight * self.weight
            
            self.weight = self.weight / torch.mean(self.weight)
        return self.weight.detach().cpu().numpy()

    def fit(self, X, y, device):
        self.device = device
        # print(X.shape)
        self.weight = self.reweight(X).squeeze()
        
        self.feature_selection_model = linear_model.LogisticRegression()
        self.feature_selection_model.fit(X,y, self.weight)
        feature_coefs = np.abs(self.feature_selection_model.coef_).reshape(-1)
        selected_index = np.where(feature_coefs>self.threshold)[0]
        self.feature_list = selected_index
        # print(self.feature_list)
        self.base_model.fit(X[:, self.feature_list], y, self.weight)
        return self.score(X,y)
        
    def score(self, X, y):
        predicted = self.base_model.predict(X[:, self.feature_list])
        correct = (predicted == y).sum()
        total = y.shape[0]
        return correct / total
    
    def f1score(self, X, y):
        predicted = self.base_model.predict(X[:, self.feature_list])
        return f1_score(y, predicted, average='macro')
    
    def predict(self, X):
        return self.base_model.predict(X[:, self.feature_list])


marginal_DRO_test_config = {
    "method": "marginal_dro",
    "device": torch.device("cuda"),
    "lr": 1e-1,
    "batch_size": 1024,
    "hidden_size": 32,
    "dropout_ratio": 0.1,
    "train_epochs": 500,
    "size": 0.4,
    "max_iter": 100,
    "radius": 1,
    "p_min": 0.1,
    "nbisect": 10
}

group_DRO_test_config = {
    "method": "marginal_dro",
    "device": torch.device("cuda"),
    "group_weights_step_size":0.01,
    "lr": 1e-1,
    "batch_size": 1024,
    "hidden_size": 32,
    "dropout_ratio": 0.1,
    "train_epochs": 50,
    "max_iter": 100,
    "size": 0.4
}
SVM_test_config = {
    "method": "svm",
    "gamma": 1,
    "C":1,
    "kernel": "rbf",
    "n_components":2000,
}


# if __name__ == "__main__":
#     # source_X = np.load('./save_data/source_X_trainincome_CA_FairThresholdOptimizer.npy').astype('float')
#     # source_y = np.load('./save_data/source_y_trainincome_CA_FairThresholdOptimizer.npy').astype('int')

#     # # method = chi_DORO(input_dim=source_X.shape[1])
#     # # method.fit(source_X, source_y)
    
#     # method = IRM(input_dim=source_X.shape[1])
#     # method.fit([source_X, source_X], [source_y, source_y])

#     source_X = np.load('./save_data/source_Xincome_CA_chi_doro.npy').astype('float')
#     source_y = np.load('./save_data/source_yincome_CA_chi_doro.npy').astype('int')
    
#     method = chi_DORO(input_dim=source_X.shape[1])
#     method.fit(source_X, source_y)
#     method.score(source_X, source_y)
#     # print(method.f1score(source_X, source_y))
#     # method = DWRPreprocess()
#     # acc = method.fit(source_X, source_y)
#     # print(acc)

#     # method = group_DRO(input_dim = source_X.shape[1])
#     # method.update(group_DRO_test_config)
#     # acc = method.fit(source_X, source_y)

#     # method = svm.SVC(cache_size = 1000, kernel = 'linear', C = 10, gamma = 2)
#     # method.fit(source_X, source_y)
#     # result = method.predict(source_X)
#     # print(accuracy_score(result, source_y))

#     # source_X = np.load('./save_data/source_X_trainincome_CA_rf.npy').astype('float')
#     # source_y = np.load('./save_data/source_y_trainincome_CA_rf.npy').astype('int')

#     # method2 = RandomForestClassifier()
#     # method2.fit(source_X, source_y)
#     # result = method2.predict(source_X)
#     # print(accuracy_score(result, source_y))
    
#     # method = SVM2()
#     # method.update(SVM_test_config)
#     # method.fit(X = source_X, y = source_y, device = device)
#     # result = method.predict(source_X)
#     # print(accuracy_score(result, source_y))

#     # method = svm.LinearSVC()
#     # from sklearn.kernel_approximation import RBFSampler
#     # rbf_feature = RBFSampler(gamma=1, random_state=1, n_components = 2000)
#     # feature_X = rbf_feature.fit_transform(source_X)
#     # method.fit(X = feature_X, y = source_y)
#     # result = method.predict(feature_X)
#     # print(accuracy_score(result, source_y))

#     # method = chi_DORO(input_dim=source_X.shape[1])
#     # method.fit(source_X, source_y)
    
#     # method = IGA(input_dim=source_X.shape[1])
#     # method.fit([source_X, source_X], [source_y, source_y])


"""Data source and problem definitions for American Community Survey (ACS) Public Use Microdata Sample (PUMS)."""
import numpy as np
import pandas as pd

from . import folktables
from .load_acs import load_acs, load_definitions


class ACSDataSource(folktables.DataSource):
    """Data source implementation for ACS PUMS data."""

    def __init__(self, survey_year, horizon, survey, root_dir="data"):
        """Create data source around PUMS data for specific year, time horizon, survey type.

        Args:
            survey_year: String. Year of ACS PUMS data, e.g., '2018'
            horizon: String. Must be '1-Year' or '5-Year'
            survey: String. Must be 'person' or 'household'

        Returns:
            ACSDataSource
        """
        if horizon not in ['1-Year', '5-Year']:
            raise ValueError(f'Horizon must be either "1-Year" or "5-Year"')
        self._survey_year = survey_year
        self._horizon = horizon
        self._survey = survey
        self._root_dir = root_dir

    def get_data(self, states=None, density=1.0, random_seed=0, join_household=False, download=False):
        """Get data from given list of states, density, and random seed. Optionally add household features."""
        data = load_acs(root_dir=self._root_dir,
                        year=self._survey_year,
                        states=states,
                        horizon=self._horizon,
                        survey=self._survey,
                        density=density,
                        random_seed=random_seed,
                        download=download)
        if join_household:
            orig_len = len(data)
            assert self._survey == 'person'
            household_data = load_acs(root_dir=self._root_dir,
                                      year=self._survey_year,
                                      states=states,
                                      horizon=self._horizon,
                                      survey='household',
                                      serial_filter_list=list(data['SERIALNO']),
                                      download=download)

            # We only want to keep the columns in the household dataframe that don't appear in the person
            # dataframe, but we *do* want to include the SERIALNO column to merge on.
            household_cols = (set(household_data.columns) - set(data.columns)).union(set(['SERIALNO']))
            join = pd.merge(data, household_data[list(household_cols)], on=['SERIALNO'])
            assert len(join) == orig_len, f'Lengths do not match after join: {len(join)} vs {orig_len}'
            return join
        else:
            return data

    def get_definitions(self, download=False):
        """
        Gets categorical data definitions dataframe.
        Only works for year>=2017 as previous years don't include .csv definition files.
        """
        return load_definitions(root_dir=self._root_dir, year=self._survey_year, horizon=self._horizon,
                                download=download)


def adult_filter(data):
    """Mimic the filters in place for Adult data.

    Adult documentation notes: Extraction was done by Barry Becker from
    the 1994 Census database. A set of reasonably clean records was extracted
    using the following conditions:
    ((AAGE>16) && (AGI>100) && (AFNLWGT>1)&& (HRSWK>0))
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['PINCP'] > 100]
    df = df[df['WKHP'] > 0]
    df = df[df['PWGTP'] >= 1]
    return df

ACSIncome = folktables.BasicProblem(
    features=[
        'AGEP',
        'COW',
        'SCHL',
        'MAR',
        'OCCP',
        'POBP',
        'RELP',
        'WKHP',
        'SEX',
        'RAC1P',
    ],
    target='PINCP',
    target_transform=lambda x: x > 50000,
    group='RAC1P',
    preprocess=adult_filter,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

ACSEmployment = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'RELP',
        'DIS',
        'ESP',
        'CIT',
        'MIG',
        'MIL',
        'ANC',
        'NATIVITY',
        'DEAR',
        'DEYE',
        'DREM',
        'SEX',
        'RAC1P',
    ],
    target='ESR',
    target_transform=lambda x: x == 1,
    group='RAC1P',
    preprocess=lambda x: x,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

ACSHealthInsurance = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'CIT',
        'MIG',
        'MIL',
        'ANC',
        'NATIVITY',
        'DEAR',
        'DEYE',
        'DREM',
        'RACAIAN',
        'RACASN',
        'RACBLK',
        'RACNH',
        'RACPI',
        'RACSOR',
        'RACWHT',
        'PINCP',
        'ESR',
        'ST',
        'FER',
    ],
    target='HINS2',
    target_transform=lambda x: x == 1,
    group='RAC1P',
    preprocess=lambda x: x,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

def public_coverage_filter(data):
    """
    Filters for the public health insurance prediction task; focus on low income Americans, and those not eligible for Medicare
    """
    df = data
    df = df[df['AGEP'] < 65]
    df = df[df['PINCP'] <= 30000]
    return df

ACSPublicCoverage = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'CIT',
        'MIG',
        'MIL',
        'ANC',
        'NATIVITY',
        'DEAR',
        'DEYE',
        'DREM',
        'PINCP',
        'ESR',
        'ST',
        'FER',
        'RAC1P',
    ],
    target='PUBCOV',
    target_transform=lambda x: x == 1,
    group='RAC1P',
    preprocess=public_coverage_filter,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

def travel_time_filter(data):
    """
    Filters for the employment prediction task
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['PWGTP'] >= 1]
    df = df[df['ESR'] == 1]
    return df

ACSTravelTime = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'MIG',
        'RELP',
        'RAC1P',
        'PUMA',
        'ST',
        'CIT',
        'OCCP',
        'JWTR',
        'POWPUMA',
        'POVPIP',
    ],
    target="JWMNP",
    target_transform=lambda x: x > 20,
    group='RAC1P',
    preprocess=travel_time_filter,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

ACSMobility = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'CIT',
        'MIL',
        'ANC',
        'NATIVITY',
        'RELP',
        'DEAR',
        'DEYE',
        'DREM',
        'RAC1P',
        'GCL',
        'COW',
        'ESR',
        'WKHP',
        'JWMNP',
        'PINCP',
    ],
    target="MIG",
    target_transform=lambda x: x == 1,
    group='RAC1P',
    preprocess=lambda x: x.drop(x.loc[(x['AGEP'] <= 18) | (x['AGEP'] >= 35)].index),
    postprocess=lambda x: np.nan_to_num(x, -1),
)

def employment_filter(data):
    """
    Filters for the employment prediction task
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['AGEP'] < 90]
    df = df[df['PWGTP'] >= 1]
    return df

ACSEmploymentFiltered = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'MIG',
        'CIT',
        'MIL',
        'ANC',
        'NATIVITY',
        'RELP',
        'DEAR',
        'DEYE',
        'DREM',
        'RAC1P',
        'GCL',
    ],
    target="ESR",
    target_transform=lambda x: x == 1,
    group='RAC1P',
    preprocess=employment_filter,
    postprocess=lambda x: np.nan_to_num(x, -1),
)

ACSIncomePovertyRatio = folktables.BasicProblem(
    features=[
        'AGEP',
        'SCHL',
        'MAR',
        'SEX',
        'DIS',
        'ESP',
        'MIG',
        'CIT',
        'MIL',
        'ANC',
        'NATIVITY',
        'RELP',
        'DEAR',
        'DEYE',
        'DREM',
        'RAC1P',
        'GCL',
        'ESR',
        'OCCP',
        'WKHP',
    ],
    target='POVPIP',
    target_transform=lambda x: x < 250,
    group='RAC1P',
    preprocess=lambda x: x,
    postprocess=lambda x: np.nan_to_num(x, -1),
)



# implement DISDE method 
# official code for DISDE could be found at https://github.com/namkoong-lab/disde

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from xgboost import XGBClassifier
from sklearn.model_selection import KFold
import random



def plot_calibration(prop_p, prop_q, nbins=20, p_weights=None, q_weights=None, 
                     nanmask_threshold=0.01, name='Prop Score',
                     save_dir='.', balance=False):
    
    fig,ax=plt.subplots(1,3,figsize=(10,4))
    for i in range(3):
        ax[i].set_box_aspect(1)
        ax[i].set_xlim(0,1)
    fig.suptitle("Calibration: {}".format(name), fontsize="x-large")

    if p_weights is None: p_weights = np.ones_like(prop_p)
    if q_weights is None: q_weights = np.ones_like(prop_q)

    p_sample_weights = p_weights.copy()
    q_sample_weights = q_weights.copy()
    if balance:
        p_sample_weights = p_sample_weights / p_sample_weights.sum()
        q_sample_weights = q_sample_weights / q_sample_weights.sum()

    conf_scores, bin_edges = np.histogram(np.concatenate([1-prop_p, prop_q]),bins=nbins, density=True, 
                                          weights=np.concatenate([p_sample_weights, 
                                                                 q_sample_weights]),
                                         range=(0,1))
    bin_mids = (bin_edges[1:]+bin_edges[:-1])/2

    nanmask = np.where(conf_scores < nanmask_threshold, np.nan, 1)
    # print(bin_mids)
    # print(nanmask * conf_scores / (conf_scores + conf_scores[::-1]))
    ax[0].plot(bin_mids, nanmask * conf_scores / (conf_scores + conf_scores[::-1]), color='green')
    ax[0].set_ylim(0,1)
    ax[0].set_ylabel('Proportion correct')
    ax[0].set_xlabel('Predicted probability')
    ax[0].set_title('Prop calibration: combined')

    conf_scores, bin_edges = np.histogram(np.concatenate([prop_p]),bins=nbins, weights=p_sample_weights,
                                         range=(0,1))
    bin_mids = (bin_edges[1:]+bin_edges[:-1])/2
    nanmask = np.where(conf_scores < nanmask_threshold, np.nan, 1)

    ax[1].plot(bin_mids, conf_scores, color='green')
    ax[1].set_title('Density: P')
    ax[1].set_xlabel('Predicted probability of Q')
    ax[1].set_ylim(bottom=0)

    conf_scores, bin_edges = np.histogram(np.concatenate([prop_q]),bins=nbins, weights=q_sample_weights,
                                         range=(0,1))
    bin_mids = (bin_edges[1:]+bin_edges[:-1])/2
    nanmask = np.where(conf_scores < nanmask_threshold, np.nan, 1)

    ax[2].plot(bin_mids, conf_scores, color='green')
    ax[2].set_title('Density: Q')
    ax[2].set_xlabel('Predicted probability of Q')
    ax[2].set_ylim(bottom=0)

    fig.tight_layout()
    if save_dir is not None:
        plt.savefig(save_dir)

def degradation_decomp(source_X, source_y, other_X_raw, other_y_raw, best_method, data_sum=20000, K=8, domain_classifier=None, draw_calibration=False, save_calibration_png='calibration.png'):
    perm1 = np.random.permutation(other_X_raw.shape[0])
    other_X = other_X_raw[perm1[:data_sum],:]
    other_y = other_y_raw[perm1[:data_sum]]

    piA = np.zeros(source_X.shape[0])
    piB = np.zeros(other_X.shape[0])
    permA = np.random.permutation(piA.shape[0])
    permB = np.random.permutation(piB.shape[0])

    kf = KFold(n_splits=K, shuffle=False)
    A_train_index_list = []    
    A_test_index_list = []    
    B_train_index_list = []    
    B_test_index_list = []    
    for i, (train_index, test_index) in enumerate(kf.split(source_X)):
        A_train_index_list.append(train_index)
        A_test_index_list.append(test_index)
    for i, (train_index, test_index) in enumerate(kf.split(other_X)):
        B_train_index_list.append(train_index)
        B_test_index_list.append(test_index)

    for i in range(K):
        trainX = np.concatenate([source_X[permA[A_train_index_list[i]]],other_X[permB[B_train_index_list[i]]]], axis=0)
        trainT = np.zeros(trainX.shape[0])
        trainT[len(A_train_index_list[i]):] = 1.0
        
        if domain_classifier is None:
            model = XGBClassifier(random_state=0).fit(trainX, trainT)
        else:
            model = domain_classifier.fit(trainX, trainT)
    
        piA[permA[A_test_index_list[i]]] = model.predict_proba(source_X[permA[A_test_index_list[i]]])[:,1]
        piB[permB[B_test_index_list[i]]] = model.predict_proba(other_X[permB[B_test_index_list[i]]])[:,1]

    if draw_calibration:
        plot_calibration(piA, piB, save_dir=save_calibration_png)

    alpha = (other_X.shape[0])/ (source_X.shape[0]+other_X.shape[0])
    wA = piA / ((1-alpha)*piA + alpha * (1-piA))
    wB = (1-piB) / ((1-alpha)*piB + alpha * (1-piB))
    accuracyA = best_method.score(source_X, source_y)
    accuracyB = best_method.score(other_X, other_y)
    wA = wA / np.sum(wA)
    wB = wB / np.sum(wB)
    predA = (best_method.predict(source_X) == source_y)
    predB = (best_method.predict(other_X) == other_y)
    sx_A = np.dot(wA, predA)
    sx_B = np.dot(wB, predB)
    return accuracyA, accuracyB, sx_A, sx_B



from typing import List
import pandas as pd
import numpy as np

from fairlearn.reductions import ExponentiatedGradient

class CustomExponentiatedGradient(ExponentiatedGradient):
    """Custom class to allow for scikit-learn-compatible interface.

    Specifically, this method takes (and ignores) a sample_weights
    parameter to its .fit() method; otherwise identical to
    fairlearn.ExponentiatedGradient.
    """

    def __init__(self, sensitive_features: List[str], **kwargs):
        super().__init__(**kwargs)
        self.sensitive_features = sensitive_features

    def fit(self, X, y, **kwargs):
        if isinstance(X, pd.DataFrame):
            super().fit(X.values, y.values,
                        sensitive_features=X[self.sensitive_features].values,
                        **kwargs)
        elif isinstance(X, np.ndarray):
            super().fit(X, y, sensitive_features = X[:, self.sensitive_features], **kwargs)
        else:
            raise NotImplementedError

    def predict_proba(self, X):
        """Alias to _pmf_predict(). Note that this tends to return 'hard'
        predictions, which don't perform well for metrics like cross-entropy."""
        return super()._pmf_predict(X)



class FolktablesException(Exception):
    """Base exception for all exceptions raised by Folktables."""


class FileDownloadError(FolktablesException):
    """Exception raised when the HTTP request to download the data fails."""


class UnsupportedOSError(FolktablesException):
    """Exception raised when the user passes the value for an OS for which
    a dataset is not supported.
    """


class InvalidFilePath(FolktablesException):
    """Exception raised when the user wants to assign the same file path to
    the data being extracted from a zip file as the zip file itself. Note
    that if we were to allow this, then the promise that the data has been
    extracted is broken as we don't want to be dealing with zip files after
    extracting the data from it.
    """


class NoFilesToDownload(FolktablesException):
    """Exception raised when there are no files to be downloaded (i.e.,
    the list is empty), but the functions to download files are still called.
    """



import argparse
import pandas as pd 
import numpy as np 
import pandas as pd 
import json
from datetime import datetime
import re
import os
import io
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import label_binarize

location_list = ["Amenity", "Bump", "Crossing", "Give_Way", "Junction", "No_Exit",  'Railway', 'Roundabout', 'Station', 'Stop', 'Traffic_Calming', 'Traffic_Signal']

def preprocess(dir):
    try:
        data = pd.read_csv(dir)
    except:
        raise FileNotFoundError('File does not exist: {}'.format(dir))
    X = data
    X["Start_Time"] = pd.to_datetime(X["Start_Time"])
    # Extract year, month, weekday and day
    X["Year"] = X["Start_Time"].dt.year
    X["Month"] = X["Start_Time"].dt.month
    X["Weekday"] = X["Start_Time"].dt.weekday
    X["Day"] = X["Start_Time"].dt.day
    # Extract hour and minute
    X["Hour"] = X["Start_Time"].dt.hour
    X["Minute"] = X["Start_Time"].dt.minute
    features_to_drop = ["ID", "Start_Time", "End_Time", "End_Lat", "End_Lng", "Description", "Number", "Street", "County", "Zipcode", "City", "Country", "Timezone", "Airport_Code", "Weather_Timestamp", "Wind_Chill(F)", "Turning_Loop", "Sunrise_Sunset", "Nautical_Twilight", "Astronomical_Twilight"]
    X = X.drop(features_to_drop, axis=1)
    X.drop_duplicates(inplace=True)
    X = X[X["Side"] != " "]
    X = X[X["Pressure(in)"] != 0]
    X = X[X["Visibility(mi)"] != 0]
    X.loc[X["Weather_Condition"].str.contains("Thunder|T-Storm", na=False), "Weather_Condition"] = "Thunderstorm"
    X.loc[X["Weather_Condition"].str.contains("Snow|Sleet|Wintry", na=False), "Weather_Condition"] = "Snow"
    X.loc[X["Weather_Condition"].str.contains("Rain|Drizzle|Shower", na=False), "Weather_Condition"] = "Rain"
    X.loc[X["Weather_Condition"].str.contains("Wind|Squalls", na=False), "Weather_Condition"] = "Windy"
    X.loc[X["Weather_Condition"].str.contains("Hail|Pellets", na=False), "Weather_Condition"] = "Hail"
    X.loc[X["Weather_Condition"].str.contains("Fair", na=False), "Weather_Condition"] = "Clear"
    X.loc[X["Weather_Condition"].str.contains("Cloud|Overcast", na=False), "Weather_Condition"] = "Cloudy"
    X.loc[X["Weather_Condition"].str.contains("Mist|Haze|Fog", na=False), "Weather_Condition"] = "Fog"
    X.loc[X["Weather_Condition"].str.contains("Sand|Dust", na=False), "Weather_Condition"] = "Sand"
    X.loc[X["Weather_Condition"].str.contains("Smoke|Volcanic Ash", na=False), "Weather_Condition"] = "Smoke"
    X.loc[X["Weather_Condition"].str.contains("N/A Precipitation", na=False), "Weather_Condition"] = np.nan
    X.loc[X["Wind_Direction"] == "CALM", "Wind_Direction"] = "Calm"
    X.loc[X["Wind_Direction"] == "VAR", "Wind_Direction"] = "Variable"
    X.loc[X["Wind_Direction"] == "East", "Wind_Direction"] = "E"
    X.loc[X["Wind_Direction"] == "North", "Wind_Direction"] = "N"
    X.loc[X["Wind_Direction"] == "South", "Wind_Direction"] = "S"
    X.loc[X["Wind_Direction"] == "West", "Wind_Direction"] = "W"
    features_to_fill = ["Temperature(F)", "Humidity(%)", "Pressure(in)", "Visibility(mi)", "Wind_Speed(mph)", "Precipitation(in)"]
    X[features_to_fill] = X[features_to_fill].fillna(X[features_to_fill].mean())
    X.dropna(inplace=True)
    X["Wind_Direction"] = X["Wind_Direction"].map(lambda x : x if len(x) != 3 else x[1:], na_action="ignore")
    X = X[X["Severity"] != 1]
    X = X[X["Severity"] != 4]
    size = len(X[X["Severity"]==3].index)
    df = pd.DataFrame()
    for i in [2,3]:
        S = X[X["Severity"]==i]
        df = df.append(S.sample(size, random_state=42))
    X = df
    X["Severity"] -= 2
    
    scaler = MinMaxScaler()
    features = ['Temperature(F)','Distance(mi)','Humidity(%)','Pressure(in)','Visibility(mi)','Wind_Speed(mph)','Precipitation(in)','Start_Lng','Start_Lat','Year', 'Month','Weekday','Day','Hour','Minute']
    X[features] = scaler.fit_transform(X[features])
    categorical_features = set(["Side", "Wind_Direction", "Weather_Condition", "Civil_Twilight"])
    
    for cat in categorical_features:
        X[cat] = X[cat].astype("category")
    X = X.replace([True, False], [1, 0])
    onehot_cols = categorical_features 
    X = pd.get_dummies(X, columns=onehot_cols, drop_first=True)
    return X 
  
    
def get_USACCdata(dir, state):
    X = preprocess(dir)
    X = X[X["State"]==state]
    sample = X
    X.to_csv("./datasets/USAccident/%s.csv"%state)
    y_sample = sample["Severity"]
    X_sample = sample.drop(["Severity", "State", "Start_Lat", "Start_Lng"], axis=1).values
    return X_sample, y_sample


"""Load ACS PUMS data from Census CSV files."""
import os
import random
import io
import requests
import zipfile

import numpy as np
import pandas as pd


state_list = ['AL', 'AK', 'AZ', 'AR', 'CA', 'CO', 'CT', 'DE', 'FL', 'GA', 'HI',
              'ID', 'IL', 'IN', 'IA', 'KS', 'KY', 'LA', 'ME', 'MD', 'MA', 'MI',
              'MN', 'MS', 'MO', 'MT', 'NE', 'NV', 'NH', 'NJ', 'NM', 'NY', 'NC',
              'ND', 'OH', 'OK', 'OR', 'PA', 'RI', 'SC', 'SD', 'TN', 'TX', 'UT',
              'VT', 'VA', 'WA', 'WV', 'WI', 'WY', 'PR']


_STATE_CODES = {'AL': '01', 'AK': '02', 'AZ': '04', 'AR': '05', 'CA': '06',
                'CO': '08', 'CT': '09', 'DE': '10', 'FL': '12', 'GA': '13',
                'HI': '15', 'ID': '16', 'IL': '17', 'IN': '18', 'IA': '19',
                'KS': '20', 'KY': '21', 'LA': '22', 'ME': '23', 'MD': '24',
                'MA': '25', 'MI': '26', 'MN': '27', 'MS': '28', 'MO': '29',
                'MT': '30', 'NE': '31', 'NV': '32', 'NH': '33', 'NJ': '34',
                'NM': '35', 'NY': '36', 'NC': '37', 'ND': '38', 'OH': '39',
                'OK': '40', 'OR': '41', 'PA': '42', 'RI': '44', 'SC': '45',
                'SD': '46', 'TN': '47', 'TX': '48', 'UT': '49', 'VT': '50',
                'VA': '51', 'WA': '53', 'WV': '54', 'WI': '55', 'WY': '56',
                'PR': '72'}


def download_and_extract(url, datadir, remote_fname, file_name, delete_download=False):
    """Helper function to download and unzip files."""
    download_path = os.path.join(datadir, remote_fname)
    response = requests.get(url)
    with open(download_path, 'wb') as handle:
        handle.write(response.content)
    
    with zipfile.ZipFile(download_path, 'r') as zip_ref:
        zip_ref.extract(file_name, path=datadir)
    
    if delete_download and download_path != os.path.join(datadir, file_name):
        os.remove(download_path)


def initialize_and_download(datadir, state, year, horizon, survey, download=False):
    """Download the dataset (if required)."""
    assert horizon in ['1-Year', '5-Year']
    # assert int(year) >= 2014
    assert state in state_list
    assert survey in ['person', 'household']

    state_code = _STATE_CODES[state]
    survey_code = 'p' if survey == 'person' else 'h'
    if int(year) >= 2017:
        file_name = f'psam_{survey_code}{state_code}.csv'
    else:
        # 2016 and earlier use different file names
        file_name = f'ss{str(year)[-2:]}{survey_code}{state.lower()}.csv'
    
    # Assume is the path exists and is a file, then it has been downloaded
    # correctly
    file_path = os.path.join(datadir, file_name)
    if os.path.isfile(file_path):
        return file_path
    if not download:
        raise FileNotFoundError(f'Could not find {year} {horizon} {survey} survey data for {state} in {datadir}. Call get_data with download=True to download the dataset.')
    
    print(f'Downloading data for {year} {horizon} {survey} survey for {state}...')
    # Download and extract file
    base_url= f'https://www2.census.gov/programs-surveys/acs/data/pums/{year}/{horizon}'
    remote_fname = f'csv_{survey_code}{state.lower()}.zip'
    url = f'{base_url}/{remote_fname}'
    try:
        download_and_extract(url, datadir, remote_fname, file_name, delete_download=True)
    except Exception as e:
        print(f'\n{os.path.join(datadir, remote_fname)} may be corrupted. Please try deleting it and rerunning this command.\n')
        print(f'Exception: ', e)

    return file_path


def load_acs(root_dir, states=None, year=2018, horizon='1-Year',
             survey='person', density=1, random_seed=1,
             serial_filter_list=None,
             download=False):
    """
    Load sample of ACS PUMS data from Census csv files into DataFrame.

    If a serial filter list is passed in, density and random_seed are ignored
    and the output is instead filtered with the provided list (only entries with
    a serial number in the list are kept).
    """
    # if int(year) < 2014:
    #     raise ValueError('Year must be >= 2014')

    if serial_filter_list is not None:
        serial_filter_list = set(serial_filter_list)  # set for faster membership check

    if states is None:
        states = state_list
    
    random.seed(random_seed)
    
    base_datadir = os.path.join(root_dir, str(year), horizon)
    os.makedirs(base_datadir, exist_ok=True)
    
    file_names = []
    for state in states:
        file_names.append(
            initialize_and_download(base_datadir, state, year, horizon, survey, download=download)
        )

    sample = io.StringIO()

    first = True
    
    for file_name in file_names:
      
        with open(file_name, 'r') as f:
            
            if first:
                sample.write(next(f))
                first = False
            else:
                next(f)

            if serial_filter_list is None:
                for line in f:
                    if random.uniform(0, 1) < density:
                        # strip whitespace found in some early files
                        sample.write(line.replace(' ',''))
            else:
                for line in f:
                    serialno = line.split(',')[1]
                    if serialno in serial_filter_list:
                        # strip whitespace found in some early files
                        sample.write(line.replace(' ',''))

            
    sample.seek(0)
    
    dtypes = {'PINCP' : np.float64, 'RT' : str, 'SOCP' : str, 'SERIALNO' : str, 'NAICSP' : str}
                    
    return pd.read_csv(sample, dtype=dtypes)


def load_definitions(root_dir, year=2018, horizon='1-Year', download=False):
    """
    Loads the data attribute definition file.

    File only available for year >= 2017.
    """
    assert horizon in ['1-Year', '5-Year']
    assert int(year) >= 2017

    base_datadir = os.path.join(root_dir, str(year), horizon)
    file_path = os.path.join(base_datadir, 'definition.csv')
    if os.path.exists(file_path):
        return pd.read_csv(file_path, sep=',', header=None, names=list(range(7)))
    if not download:
        raise FileNotFoundError(
            f'Could not find {year} {horizon} attribute definition. Call get_definitions with download=True to download the definitions.')

    # download definition first
    print('Downloading the attribute definition file...')
    year_string = year if horizon == '1-Year' else f'{year - 4}-{year}'
    url = f'https://www2.census.gov/programs-surveys/acs/tech_docs/pums/data_dict/PUMS_Data_Dictionary_{year_string}.csv'

    response = requests.get(url)
    with open(file_path, 'wb') as handle:
        handle.write(response.content)

    return pd.read_csv(file_path, sep=',', header=None, names=list(range(7)))


def generate_categories(features, definition_df):
    """Generates a categories dictionary using the provided definition dataframe. Does not create a category mapping
    for variables requiring the 2010 Public use microdata area code (PUMA) as these need an additional definition
    file which are not unique without the state code.

    Args:
        features: list (list of features to include in the categories dictionary, numeric features will be ignored)
        definition_df: pd.DataFrame (received from ```ACSDataSource.get_definitions()''')

    Returns:
        categories: nested dict with columns of categorical features
            and their corresponding encodings (see examples folder)."""
    categories = {}
    for feature in features:
        if 'PUMA' in feature:
            continue

        # extract definitions for this feature
        coll_definition = definition_df[(definition_df[0] == 'VAL') & (definition_df[1] == feature)]

        # extracts if the feature is numeric or categorical --> 'N' == numeric
        coll_type = coll_definition.iloc[0][2]
        if coll_type == 'N':
            # do not add to categories
            continue

        # transform to numbers as downloaded definitions are in string format.
        # -99999999999999.0 is used as a placeholder value for NaN
        # as multiple NaN values are seen as different keys in a dictionary, a placeholder is needed
        mapped_col = pd.to_numeric(coll_definition[4], errors='coerce').fillna(-99999999999999.0)
        mapping_dict = dict(zip(mapped_col.tolist(), coll_definition[6].tolist()))

        # add default value when not already available from definitions
        if -99999999999999.0 not in mapping_dict:
            mapping_dict[-99999999999999.0] = 'N/A'
        # transform placeholder value back to NaN ensuring a single NaN key instaid of multiple
        mapping_dict[float('nan')] = mapping_dict[-99999999999999.0]
        del mapping_dict[-99999999999999.0]

        categories[feature] = mapping_dict
    return categories


"""Implements abstract classes for folktables data source and problem definitions."""

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd


class DataSource(ABC):
    """Provides access to data source."""

    @abstractmethod
    def get_data(self, **kwargs):
        """Get data sample from universe.

        Returns:
            Sample."""
        pass


class Problem(ABC):
    """Abstract class for specifying learning problem."""

    @abstractmethod
    def df_to_numpy(self, df):
        """Return learning problem as numpy array."""
        pass

    # Returns the column name
    @property    
    @abstractmethod
    def target(self):
        pass

    @property
    @abstractmethod
    def features(self):
        pass

    @property
    @abstractmethod
    def target_transform(self):
        pass


class BasicProblem(Problem):
    """Basic prediction or regression problem."""

    def __init__(self,
                 features,
                 target,
                 target_transform=None,
                 group=None,
                 group_transform=lambda x: x,
                 preprocess=lambda x: x,
                 postprocess=lambda x: x):
        """Initialize BasicProblem.

        Args:
            features: list of column names to use as features
            target: column name of target variable
            target_transform: feature transformation for target variable
            group: designated group membership feature
            group_transform: feature transform for group membership
            preprocess: function applied to initial data frame
            postprocess: function applied to final numpy data array
        """
        self._features = features
        self._target = target
        self._target_transform = target_transform
        self._group = group
        self._group_transform = group_transform
        self._preprocess = preprocess
        self._postprocess = postprocess

    def df_to_numpy(self, df):
        """Return data frame as numpy array.
        
        Args:
            DataFrame.
        
        Returns:
            Numpy array, numpy array, numpy array"""

        df = self._preprocess(df)
        res = []
        for feature in self.features:
            res.append(df[feature].to_numpy())
        res_array = np.column_stack(res)
        
        if self.target_transform is None:
            target = df[self.target].to_numpy()
        else:
            target = self.target_transform(df[self.target]).to_numpy()
        
        if self._group:
            group = self.group_transform(df[self.group]).to_numpy()
        else:
            group = np.zeros(len(target))

        return self._postprocess(res_array), target, group

    def df_to_pandas(self, df, categories=None, dummies=False):
        """Filters and processes a DataFrame (received from ```ACSDataSource''').
        
        Args:
            df: pd.DataFrame (received from ```ACSDataSource''')
            categories: nested dict with columns of categorical features
                and their corresponding encodings (see examples folder)
            dummies: bool to indicate the creation of dummy variables for
                categorical features (see examples folder)
        
        Returns:
            pandas.DataFrame."""
        
        df = self._preprocess(df)

        variables = df[self.features]

        if categories:
            variables = variables.replace(categories)
        
        if dummies:
            variables = pd.get_dummies(variables)

        variables = pd.DataFrame(self._postprocess(variables.to_numpy()),
                                 columns=variables.columns)

        if self.target_transform is None:
            target = df[self.target]
        else:
            target = self.target_transform(df[self.target])

        target = pd.DataFrame(target).reset_index(drop=True)

        if self._group:
            group = self.group_transform(df[self.group])
            group = pd.DataFrame(group).reset_index(drop=True)
        else:
            group = pd.DataFrame(0, index=np.arange(len(target)), columns=["group"])

        return variables, target, group

    @property
    def target(self):
        return self._target
    
    @property
    def target_transform(self):
        return self._target_transform
    
    @property
    def features(self):
        return self._features
    
    @property
    def group(self):
        return self._group

    @property
    def group_transform(self):
        return self._group_transform



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.model_selection import KFold
from sklearn.tree import DecisionTreeRegressor
import sklearn 
import random
import graphviz 
import xgboost as xgb 
from .dataset import get_data


def shared_reweight(source_X, other_X, K=8):
    piA = np.zeros(source_X.shape[0])
    piB = np.zeros(other_X.shape[0])
    permA = np.random.permutation(piA.shape[0])
    permB = np.random.permutation(piB.shape[0])
    kf = KFold(n_splits=K, shuffle=False)
    A_train_index_list = []    
    A_test_index_list = []    
    B_train_index_list = []    
    B_test_index_list = []    
    for i, (train_index, test_index) in enumerate(kf.split(source_X)):
        A_train_index_list.append(train_index)
        A_test_index_list.append(test_index)
    for i, (train_index, test_index) in enumerate(kf.split(other_X)):
        B_train_index_list.append(train_index)
        B_test_index_list.append(test_index)

    for i in range(K):
        trainX = np.concatenate([source_X[permA[A_train_index_list[i]]],other_X[permB[B_train_index_list[i]]]], axis=0)
        trainT = np.zeros(trainX.shape[0])
        trainT[len(A_train_index_list[i]):] = 1.0
        model = xgb.XGBClassifier(random_state=0)
        model.fit(trainX,trainT)
        piA[permA[A_test_index_list[i]]] = model.predict_proba(source_X[permA[A_test_index_list[i]]])[:,1]
        piB[permB[B_test_index_list[i]]] = model.predict_proba(other_X[permB[B_test_index_list[i]]])[:,1]
    
    alpha = (other_X.shape[0])/ (source_X.shape[0]+other_X.shape[0])
    wA = piA / ((1-alpha)*piA + alpha * (1-piA))
    wB = (1-piB) / ((1-alpha)*piB + alpha * (1-piB))
    wA = wA * 10000 / wA.sum()
    wB = wB * 10000 / wB.sum()
    
    new_X = np.concatenate([source_X, other_X], axis=0)
    new_weights = np.concatenate([wA,wB])
    new_weights /= np.sum(new_weights)

    return wA, wB, new_X, new_weights


def risk_region(method, source_model, target_model, task, source_state, target_state, root_dir='./dataset/acs/', need_preprocess=False, year = 2018):
    num_samples = 10000
    test_results = {}
    test_X = {}
    test_y = {}
    
    sample_sizes = [20000, 10000]
    for idx, state in enumerate([source_state, target_state]):
        source_X_raw, source_y_raw, feature_names = get_data(task, state, need_preprocess, root_dir, year)

        if source_X_raw.shape[0]>=2*num_samples:
            perm2 = np.random.permutation(source_X_raw.shape[0])[:sample_sizes[idx]]
            test_X[state] = source_X_raw[perm2,:]
            test_y[state] = source_y_raw[perm2]
        else:
            perm2 = np.random.permutation(source_X_raw.shape[0])[:sample_sizes[idx]]
            test_X[state] = source_X_raw[perm2,:]
            test_y[state] = source_y_raw[perm2]
            

    wA, wB, new_X, new_weights = shared_reweight(test_X[source_state], test_X[target_state])
    
    if not method == 'mlp':
        source_model.fit(test_X[source_state], test_y[source_state], sample_weight = wA)
        target_model.fit(test_X[target_state], test_y[target_state], sample_weight = wB)    
    else:
        print(wA.shape, wB.shape)
        source_model.fit_weight(test_X[source_state], test_y[source_state], wA.reshape(-1))
        target_model.fit_weight(test_X[target_state], test_y[target_state],wB.reshape(-1))    

    source_acc = source_model.score(test_X[source_state], test_y[source_state])
    target_acc = source_model.score(test_X[target_state], test_y[target_state])

    print("source model state %s acc is %.4f" % (source_state, source_acc))
    print("source model state %s acc is %.4f" % (target_state, target_acc))   

    source_acc2 = target_model.score(test_X[source_state], test_y[source_state])
    target_acc2 = target_model.score(test_X[target_state], test_y[target_state])
    print("tgt model state %s acc is %.4f" % (source_state, source_acc2))
    print("tgt model state %s acc is %.4f" % (target_state, target_acc2))   



    eps = 1e-5
    new_X = np.concatenate([test_X[source_state], test_X[target_state]], axis = 0)
    proba_P2Q = np.clip(source_model.predict_proba(new_X)[:, -1], eps, 1 - eps)
    proba_Q2P = np.clip(target_model.predict_proba(new_X)[:, -1], eps, 1 - eps)
    new_Y = np.abs(proba_P2Q - proba_Q2P)

    plt.figure(figsize=(50,50), dpi=200)
    region_model = DecisionTreeRegressor(max_depth=6,min_samples_leaf=100,min_samples_split=200,min_weight_fraction_leaf=0.05, ccp_alpha=0.0001).fit(new_X, new_Y, sample_weight=new_weights)
    sklearn.tree.plot_tree(region_model, filled=True, feature_names = feature_names)
    plt.savefig(f'{method}_{task}_{source_state}to{target_state}_region.pdf')


  




import pathlib
import zipfile

from folktables import exceptions


def extract_content_from_zip(data_dir, file_name, zip_file):
    """Extracts the contents from a Zip file, and removes the Zip file
    if its file path is not the same as that of the extracted
    file.

    Parameters
    ----------
    data_dir : str
        Path to the directory where the file will be stored.
    file_name : str
        Name that will be given to the extracted file.
    zip_file : pathlib.Path
        Path to where the Zip file is located.
    """
    if not isinstance(zip_file, pathlib.Path):
        # We want to make sure that even if the user doesn't pass a
        # pathlib.Path object, that this function doesn't break because
        # of that.
        zip_file = pathlib.Path(zip_file)

    if zip_file == pathlib.Path(data_dir, file_name):
        # We want to avoid the case in which the download path
        # (i.e., the zip asset) has the same name as the file to be extracted.
        raise exceptions.InvalidFilePath(
           f'Invalid `data_dir` and `file_name`. The path resolves to:\n'
           f'{pathlib.Path(data_dir, file_name).resolve()}'
           f'Please make sure that the above path is not the same as that '
           f'to which the data was downloaded:\n{zip_file.resolve()}'
        )

    with zipfile.ZipFile(zip_file, 'r') as zip_ref:
        zip_ref.extract(file_name, path=data_dir)

    zip_file.unlink()



from sklearn.svm import SVC
import numpy as np 
import os
from sklearn.ensemble import GradientBoostingClassifier, RandomForestClassifier, RandomForestRegressor
from lightgbm import LGBMClassifier
import xgboost as xgb
from .methods.model_family import *


def fetch_model(method_name, input_dim=10, num_classes=2):
    if method_name == 'lr':
        return LogisticRegression(input_dim, num_classes)
    elif method_name == 'mlp':
        return MLPClassifier(input_dim, num_classes)
    elif method_name == 'chi_dro':
        return chi_DRO(input_dim, num_classes)
    elif method_name == 'cvar_dro':
        return cvar_DRO(input_dim, num_classes)
    elif method_name == 'cvar_doro':
        return cvar_DORO(input_dim, num_classes)
    elif method_name == 'chi_doro':
        return chi_DORO(input_dim, num_classes)
    elif method_name == 'group_dro':
        return group_DRO(input_dim, num_classes)
    elif method_name == 'marginal_dro':
        return marginal_DRO(input_dim, num_classes)
    elif method_name in ['FairPostprocess', 'FairPostprocess_exp','FairPostprocess_threshold']:
        return FairPostprocess()
    elif method_name in ['FairInprocess', 'FairInprocess_eo', 'FairInprocess_dp', 'FairInprocess_error_parity']:
        return FairInprocess()
    elif method_name == 'jtt':
        return JTT()
    elif method_name == 'svm2':
        return SVM2()
    elif method_name == 'dwr':
        return DWRPreprocess()
    elif method_name == 'rf':
        return RandomForestClassifier()
    elif method_name == 'gbm':
        return GradientBoostingClassifier()
    elif method_name == 'lightgbm':
        return LGBMClassifier()
    elif method_name == 'svm':
        return SVC(cache_size=1000, max_iter=1000)
    elif method_name in ['xgb', 'subg', 'rwy', 'rwg', 'suby']:
        return xgb.XGBClassifier()
    else:
        raise NotImplementedError



import pandas as pd
import numpy as np
import functools
from sklearn.model_selection import train_test_split
from .folktables import adult_filter

def flatten_list(list_of_lists):
        return functools.reduce(lambda x,y:x+y, list_of_lists)

rs = 0
s = 'AL,AK,AZ,AR,CA,CO,CT,DE,FL,GA,HI,ID,IL,IN,IA,KS,KY,LA,ME,MD,MA,MI,MN,MS,MO,MT,NE,NV,NH,NJ,NM,NY,NC,ND,OH,OK,OR,PA,RI,SC,SD,TN,TX,UT,VT,VA,WA,WV,WI,WY'
all_states = s.split(',')
rac1p_vals = ['white','black','am_ind','alaska','am_alaska','asian','hawaiian','other','two_or_more']
feature_categories = ['AGEP','RACE','SCHL','MAR','MIL','CIT','MIG','DIS','SEX','NATIVITY']
relp_vals = ['reference', 'husband/wife','biologicalson','adoptedson','stepson','brother','father','grandchild','parentinlaw','soninlaw','other','roomer',
            'housemate','unmarried','foster','nonrelative','institutionalized','noninstitutionalized']
jwtr_vals = ['car', 'bus', 'streetcar', 'subway', 'railroad', 'ferryboat', 'taxicab', 'motocycle', 'bicycle', 'walk', 'home', 'other']
COW_vals = ['cow_employee_profit', 'cow_employee_nonprofit', 'cow_localgovernment', 'cow_stategovernment', 'cow_federalgovernment', 'cow_selfemployed_own',\
                'cow_selfemployed_incorporated', 'cow_family_business', 'cow_unemployed']
OCCP_vals = ['MGR1', 'MGR2', 'MGR3', 'MGR4', 'MGR5', 'BUS1', 'BUS2', 'BUS3', 'FIN1', 'FIN2', 
             'CMM1', 'CMM2', 'CMM3', 'ENG1', 'ENG2', 'ENG3', 'SCI1', 'SCI2', 'SCI3', 'SCI4', 
             'CMS1', 'LGL1', 'EDU1', 'EDU2', 'EDU3', 'EDU4', 'ENT1', 'ENT2', 'ENT3', 'ENT4', 
             'MED1', 'MED2', 'MED3', 'MED4', 'MED5', 'MED6', 'HLS1', 'PRT1', 'PRT2', 'PRT3',  
             'EAT1', 'EAT2', 'CLN1', 'PRS1', 'PRS2', 'PRS3', 'PRS4', 'SAL1', 'SAL2', 'SAL3', 
             'OFF1', 'OFF2', 'OFF3', 'OFF4', 'OFF5', 'OFF6', 'OFF7', 'OFF8', 'OFF9', 'OFF10',
             'FFF1', 'FFF2', 'CON1', 'CON2', 'CON3', 'CON4', 'CON5', 'CON6', 'EXT1', 'EXT2', 
             'RPR1', 'RPR2', 'RPR3', 'RPR4', 'RPR5', 'RPR6', 'RPR7', 'PRD1', 'PRD2', 'PRD3', 
             'PRD4', 'PRD5', 'PRD6', 'PRD7', 'PRD8', 'PRD9', 'PRD10', 'PRD11', 'PRD12', 'PRD13', 
             'TRN1', 'TRN2', 'TRN3', 'TRN4', 'TRN5', 'TRN6', 'TRN7', 'TRN8', 'MIL1', "no1"]

Big_OCCP_vals = ['MGR', 'BUS', 'FIN', 'CMM', 'ENG', 'SCI', 'CMS', 'LGL', 'EDU', 'ENT', 'MED', 'HLS', 'PRT', 'EAT', 'CLN', 'PRS',
                'SAL', 'OFF', 'FFF', 'CON', 'EXT', 'RPR', 'PRD', 'TRN', 'MIL', 'no']
Big_OCCP_threshold = [[0,4], [5,7], [8,9], [10,12], [13,15], [16,19], [20, 20], [21, 21], [22, 25],\
                    [26, 29], [30,35], [36,36], [37,39], [40,41], [42, 42], [43,46], [47,49], [50,59],\
                    [60,61], [62,67], [68,69], [70,76], [77,89], [90,97], [98,98],[99,99]]
Large_OCCP_vals = ['Lg', 'Semi_Lg', 'Non_Lg']
CIT_vals = ['us', 'pr', 'abroad', 'citizen', 'not']
ESR_vals = ['employed', 'partial_employed', 'unemployed', 'armed', 'partial_armed', 'no']

def add_esr_indicators(t):
    for idx, esr in enumerate(ESR_vals):
        t['ESR_%s'%esr] = t.ESR == (idx+1)

def add_cit_indicators(t):
    for idx, cit in enumerate(CIT_vals):
        t['CIT_%s'%cit] = t.CIT == (idx+1)

def add_occp_indicators(t):
    for idx, occp in enumerate(OCCP_vals):
        t['occp_%s'%occp] = t.OCCP == idx
    return t

def add_big_OCCP_indicators(t):
    for idx, big_occp in enumerate(Big_OCCP_vals):
        t['big_occp_%s'%big_occp] = ((t.OCCP >= Big_OCCP_threshold[idx][0]) & (t.OCCP <= Big_OCCP_threshold[idx][1]))
    return t 

def add_large_occp_indicator(t):
    t['large_occp_Lg'] = ((t.big_occp_MGR == 1)|(t.big_occp_BUS == 1)|(t.big_occp_FIN == 1)|\
                          (t.big_occp_LGL == 1)|(t.big_occp_EDU == 1)|(t.big_occp_ENT == 1))

    t['large_occp_Semi_Lg'] = ((t.big_occp_CMM == 1)|(t.big_occp_ENG == 1)|(t.big_occp_SCI == 1)|\
                          (t.big_occp_CMS == 1)|(t.big_occp_MED == 1)|(t.big_occp_HLS == 1)|(t.big_occp_PRS == 1)|\
                          (t.big_occp_SAL == 1)|(t.big_occp_OFF == 1)|(t.big_occp_RPR == 1)|(t.big_occp_PRD == 1))

    t['large_occp_Non_Lg'] = ((t.big_occp_PRT == 1)|(t.big_occp_EAT == 1)|(t.big_occp_CLN == 1)|\
                          (t.big_occp_FFF == 1)|(t.big_occp_CON == 1)|(t.big_occp_EXT == 1)|(t.big_occp_TRN == 1)\
                          |(t.big_occp_MIL == 1)|(t.big_occp_no == 1))


def add_race_indicators(t):
    for idx, race in enumerate(rac1p_vals):
        t['race_%s'%race] = t.RAC1P == (idx+1)
    return t
def add_relp_indicators(t):
    for idx, relp in enumerate(relp_vals):
        t['relp_%s'%relp] = t.RELP == idx
    return t

def add_jwtr_indicators(t):
    for idx, relp in enumerate(jwtr_vals):
        t['jwtr_%s'%relp] = t.JWTR == (idx+1)
    return t

def add_school_indicators(t):
    t['schl_at_least_bachelor']=t.SCHL >= 21
    t['schl_at_least_high_school_or_ged']=t.SCHL >= 17
    t['schl_postgrad']=t.SCHL >= 22
    return t
def add_married_indicator(t):
    t['married']=t.MAR==1
    t['widowed']=t.MAR==2
    t['divorced']=t.MAR==3
    t['separated']=t.MAR==4
    t['never']=t.MAR==5
    return t
def add_military_indicators(t):
    t['active_military']=t.MIL==1
    t['vet']=t.MIL==2
    return t
def add_citizenship_indicators(t):
    t['cit_born_us']=t.CIT==1
    t['cit_born_territory']=t.CIT==2
    t['cit_am_parents']=t.CIT==3
    t['cit_naturalized']=t.CIT==4
    t['cit_not_citizen']=t.CIT==5
    return t
def add_mobility_indicators(t):
    t['mig_moved']=t.MIG!=1
    return t

def add_cow_indicators(t):
    for idx, cow in enumerate(COW_vals):
        t[cow] = t.COW == (idx+1)
    return t
        


def add_indicators(t):
    adult_filter(t)
    add_race_indicators(t)
    add_relp_indicators(t)
    add_school_indicators(t)
    add_married_indicator(t)
    t.OCCP = t.OCCP // 100
    add_big_OCCP_indicators(t)
    # add_occp_indicators(t)
    add_large_occp_indicator(t)
    add_cow_indicators(t)
    return t

def add_indicators_year(t):
    adult_filter(t)
    add_race_indicators(t)
    # add_relp_indicators(t)
    add_school_indicators(t)
    add_married_indicator(t)
    
    # add_military_indicators(t)
    # add_citizenship_indicators(t)
    # add_mobility_indicators(t)
    return t

def add_indicators_pubcov(t):
    add_race_indicators(t)
    add_school_indicators(t)
    add_married_indicator(t)
    add_cit_indicators(t)
    add_esr_indicators(t)
    return t

def add_indicators_traveltime(t):
    add_race_indicators(t)
    add_school_indicators(t)
    add_married_indicator(t)
    add_jwtr_indicators(t)
    return t

def add_indicators_mobility(t):
    add_race_indicators(t)
    add_school_indicators(t)
    add_married_indicator(t)
    add_relp_indicators(t)
    add_cow_indicators(t)
    add_esr_indicators(t)
    return t

binarized_features = ['race_'+x for x in rac1p_vals] + ['married',
                                                    'schl_at_least_bachelor',
                                                    'schl_at_least_high_school_or_ged',
                                                    'schl_postgrad',
                                                    'active_military','vet',
                                                    'cit_born_us',
                                                    'cit_born_territory',
                                                    'cit_am_parents',
                                                    'cit_naturalized',
                                                    'cit_not_citizen','mig_moved']



import numpy as np
from . import folktables_utils
from .folktables import ACSDataSource, ACSEmployment, ACSIncome, ACSPublicCoverage, ACSMobility, ACSTravelTime
from .folktables import BasicProblem
import pandas as pd
from sklearn import preprocessing
import pickle
from sklearn.preprocessing import StandardScaler
from .folktables_utils import add_indicators, add_indicators_year, add_indicators_pubcov, add_indicators_traveltime, add_indicators_mobility
from .utils import preprocess

rac1p_vals = ['white','black','am_ind','alaska','am_alaska','asian','hawaiian','other','two_or_more']
relp_vals = ['reference', 'husband/wife','biologicalson','adoptedson','stepson','brother','father','grandchild','parentinlaw','soninlaw','other','roomer',
    'housemate','unmarried','foster','nonrelative','institutionalized','noninstitutionalized']
SCHL_vals = ['SCHL', 'schl_at_least_bachelor', 'schl_at_least_high_school_or_ged', 'schl_postgrad']
COW_vals = ['cow_employee_profit', 'cow_employee_nonprofit', 'cow_localgovernment', 'cow_stategovernment', 'cow_federalgovernment', 'cow_selfemployed_own',\
            'cow_selfemployed_incorporated', 'cow_family_business', 'cow_unemployed']
OCCP_vals = ['MGR1', 'MGR2', 'MGR3', 'MGR4', 'MGR5', 'BUS1', 'BUS2', 'BUS3', 'FIN1', 'FIN2', 
             'CMM1', 'CMM2', 'CMM3', 'ENG1', 'ENG2', 'ENG3', 'SCI1', 'SCI2', 'SCI3', 'SCI4', 
             'CMS1', 'LGL1', 'EDU1', 'EDU2', 'EDU3', 'EDU4', 'ENT1', 'ENT2', 'ENT3', 'ENT4', 
             'MED1', 'MED2', 'MED3', 'MED4', 'MED5', 'MED6', 'HLS1', 'PRT1', 'PRT2', 'PRT3',  
             'EAT1', 'EAT2', 'CLN1', 'PRS1', 'PRS2', 'PRS3', 'PRS4', 'SAL1', 'SAL2', 'SAL3', 
             'OFF1', 'OFF2', 'OFF3', 'OFF4', 'OFF5', 'OFF6', 'OFF7', 'OFF8', 'OFF9', 'OFF10', 
             'FFF1', 'FFF2', 'CON1', 'CON2', 'CON3', 'CON4', 'CON5', 'CON6', 'EXT1', 'EXT2', 
             'RPR1', 'RPR2', 'RPR3', 'RPR4', 'RPR5', 'RPR6', 'RPR7', 'PRD1', 'PRD2', 'PRD3', 
             'PRD4', 'PRD5', 'PRD6', 'PRD7', 'PRD8', 'PRD9', 'PRD10', 'PRD11', 'PRD12', 'PRD13', 
             'TRN1', 'TRN2', 'TRN3', 'TRN4', 'TRN5', 'TRN6', 'TRN7', 'TRN8', 'MIL1', "no1"]
Big_OCCP_vals = ['MGR', 'BUS', 'FIN', 'CMM', 'ENG', 'SCI', 'CMS', 'EDU', 'ENT', 'MED', 'HLS', 'PRT', 'EAT', 'CLN', 'PRS',\
                'SAL', 'OFF', 'FFF', 'CON', 'EXT', 'RPR', 'PRD', 'TRN', 'MIL', 'no']
Large_OCCP_vals = ['Lg', 'Semi_Lg', 'Non_Lg']
CIT_vals = ['us', 'pr', 'abroad', 'citizen', 'not']
ESR_vals = ['employed', 'partial_employed', 'unemployed', 'armed', 'partial_armed', 'no']


METHOD_NONE_DUMMIES = []
METHOD_NEED_PREPROCESS = ['svm', 'chi_doro', 'cvar_doro']

def adult_filter(data):
    """Mimic the filters in place for Adult data.
    Adult documentation notes: Extraction was done by Barry Becker from
    the 1994 Census database. A set of reasonably clean records was extracted
    using the following conditions:
    ((AAGE>16) && (AGI>100) && (AFNLWGT>1)&& (HRSWK>0))
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['PINCP'] > 100]
    df = df[df['WKHP'] > 0]
    df = df[df['PWGTP'] >= 1]
    return df
def travel_time_filter(data):
    """
    Filters for the employment prediction task
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['PWGTP'] >= 1]
    df = df[df['ESR'] == 1]
    return 

def employment_filter(data):
    """
    Filters for the employment prediction task
    """
    df = data
    df = df[df['AGEP'] > 16]
    df = df[df['AGEP'] < 90]
    df = df[df['PWGTP'] >= 1]
    return 

def public_coverage_filter(data):
    """
    Filters for the public health insurance prediction task; focus on low income Americans, and those not eligible for Medicare
    """
    df = data
    df = df[df['AGEP'] < 65]
    df = df[df['PINCP'] <= 30000]
    return df

def mobility_filter(data):
    df = data 
    df = df[df["AGEP"] > 18]
    df = df[df["AGEP"] < 35]
    return df

def get_USAccident(state, need_preprocess, root_dir='./datasets/USAccident/US_Accidents_Dec21_updated.csv'):
    if not state in ['CA', 'TX', 'FL', 'OR', 'MN', 'VA', 'SC', 'NY', 'PA', 'NC', 'TN', 'MI', 'MO']:
        raise NotImplementedError(f"{state} is not supported in this dataset!")

    raw_X = preprocess(root_dir)
    data = raw_X[raw_X["State"]==state]

    y_sample = data["Severity"]
    X_sample = data.drop(["Severity", "State", "Start_Lat", "Start_Lng"], axis=1).values

    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(X_sample)
        X_sample = scaler.transform(X_sample)

    # print(X_sample.shape, yq_sample.shape)
    return X_sample, y_sample.to_numpy().astype('int'), None

def get_taxi(city, need_preprocess,root_dir='./datasets/taxi/nyc_clean.csv'):

    remove_col_nyc = ['id', 'pickup_latitude', 'pickup_longitude', 'dropoff_latitude', 'dropoff_longitude', 'passenger_count']
    remove_col_other = ['id', 'dist_meters', 'wait_sec', 'pickup_latitude', 'pickup_longitude', 'dropoff_latitude', 'dropoff_longitude']

    try:
        df = pd.read_csv(root_dir)
    except:
        raise FileNotFoundError('File does not exist: {}'.format(root_dir))
        
    df = df[(df.trip_duration < 5900)]
    df = df[(df.pickup_longitude > -110)]
    df = df[(df.pickup_latitude < 50)]
    df.drop(['store_and_fwd_flag'], axis=1, inplace=True)
    df.drop(['vendor_id'], axis=1, inplace=True)
    df['pickup_datetime'] = pd.to_datetime(df.pickup_datetime)
    df.drop(['dropoff_datetime'], axis=1, inplace=True) 
    df['month'] = df.pickup_datetime.dt.month
    df['week'] = df.pickup_datetime.dt.isocalendar().week
    df['weekday'] = df.pickup_datetime.dt.weekday
    df['hour'] = df.pickup_datetime.dt.hour
    df['minute'] = df.pickup_datetime.dt.minute
    df['minute_oftheday'] = df['hour'] * 60 + df['minute']
    df.drop(['minute'], axis=1, inplace=True)


    df.drop(['pickup_datetime'], axis=1, inplace=True)

    def ft_haversine_distance(lat1, lng1, lat2, lng2):
        lat1, lng1, lat2, lng2 = map(np.radians, (lat1, lng1, lat2, lng2))
        AVG_EARTH_RADIUS = 6371 #km
        lat = lat2 - lat1
        lng = lng2 - lng1
        d = np.sin(lat * 0.5) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(lng * 0.5) ** 2
        h = 2 * AVG_EARTH_RADIUS * np.arcsin(np.sqrt(d))
        return h

    df['distance'] = ft_haversine_distance(df['pickup_latitude'].values,
                                                    df['pickup_longitude'].values, 
                                                    df['dropoff_latitude'].values,
                                                    df['dropoff_longitude'].values)

    def ft_degree(lat1, lng1, lat2, lng2):
        AVG_EARTH_RADIUS = 6371 #km
        lng_delta_rad = np.radians(lng2 - lng1)
        lat1, lng1, lat2, lng2 = map(np.radians, (lat1, lng1, lat2, lng2))
        y = np.sin(lng_delta_rad) * np.cos(lat2)
        x = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(lng_delta_rad)
        return np.degrees(np.arctan2(y, x))

    df['direction'] = ft_degree(df['pickup_latitude'].values,
                                    df['pickup_longitude'].values,
                                    df['dropoff_latitude'].values,
                                    df['dropoff_longitude'].values)


    df = df[(df.distance < 200)]
    df['speed'] = df.distance / df.trip_duration
    df = df[(df.speed < 30)]

    df.drop(['speed'], axis=1, inplace=True)

    try:
        df = df.sample(n = 10000)
        test = test.sample(n = 10000)
    except:
        pass

    y_sample = df["trip_duration"].apply(lambda x: 0 if x < 900 else 1)
    df.drop(["trip_duration"], axis=1, inplace=True)
    if city == 'nyc':
        df.drop(remove_col_nyc, axis=1, inplace=True)
    else:
        df.drop(remove_col_other, axis = 1, inplace = True)

    X_sample = df.to_numpy()

    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(X_sample)
        X_sample = scaler.transform(X_sample)

    return X_sample, y_sample.to_numpy().astype('int'), None

def get_ACSIncome(state, year=2018, need_preprocess=True, root_dir = './datasets/acs'):
    task = ACSIncome
    data_source = ACSDataSource(root_dir=root_dir, survey_year=year, horizon='1-Year', survey='person')
    source_data = data_source.get_data(states=[state], download=True)
    
    source_data = adult_filter(source_data)
    
    new_features = ['SEX', 'AGEP', 'WKHP',  'married', 'widowed','divorced','separated','never']+['relp_'+x for x in relp_vals]\
                    +['race_'+x for x in rac1p_vals] + COW_vals +['big_occp_'+x for x in Big_OCCP_vals]\
                    +SCHL_vals+['large_occp_'+x for x in Large_OCCP_vals]
    
    new_task = BasicProblem(new_features, task._target, task._target_transform,
                task._group, task._group_transform,
                preprocess = add_indicators, postprocess = task._postprocess)

    source_X_raw, source_y_raw, _ = new_task.df_to_numpy(source_data)
    
    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(source_X_raw[:,2:])
        source_X_raw[:,2:] = scaler.transform(source_X_raw[:,2:])
        
    return source_X_raw, source_y_raw.astype("int"), new_features

def get_ACSIncome_aug(state, year=2018, need_preprocess=True, root_dir = './datasets/acs'):
    """
    Add features to mitigate the concept drifts
    """
    task = ACSIncome
    data_source = ACSDataSource(root_dir=root_dir, survey_year=year, horizon='1-Year', survey='person')
    source_data = data_source.get_data(states=[state], download=True)
    source_data = adult_filter(source_data)
    
    new_features = ['SEX', 'AGEP', 'WKHP',  'married', 'widowed','divorced','separated','never']+['relp_'+x for x in relp_vals]\
                    +['race_'+x for x in rac1p_vals] + COW_vals +['big_occp_'+x for x in Big_OCCP_vals]\
                    +SCHL_vals+['large_occp_'+x for x in Large_OCCP_vals]+['ENG']
    
    new_task = BasicProblem(new_features, task._target, task._target_transform,
                task._group, task._group_transform,
                preprocess = add_indicators, postprocess = task._postprocess)

    source_X_raw, source_y_raw, _ = new_task.df_to_numpy(source_data)
    
    index = np.where(source_X_raw[:,-1]>2)[0]
    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(source_X_raw[:,1:])
        source_X_raw[:,1:] = scaler.transform(source_X_raw[:,1:])

    return source_X_raw, source_y_raw.astype("int"), new_features

def get_ACSPubCov(state, year=2018, need_preprocess=True, root_dir = './datasets/acs'):
    task = ACSPublicCoverage
    data_source = ACSDataSource(root_dir=root_dir, survey_year=year, horizon='1-Year', survey='person')
    source_data = data_source.get_data(states=[state], download=False)
    
    source_data = public_coverage_filter(source_data)
    rac1p_vals = ['white','black','am_ind','alaska','am_alaska','asian','hawaiian','other','two_or_more']
    
    new_features = ['SEX', 'AGEP', 'DIS', 'ESP', 'MIG', 'MIL', 'ANC', 'NATIVITY', 'DEAR', 'DEYE',
                    'DREM', 'PINCP', 'FER',  'married', 'widowed','divorced','separated','never']+['race_'+x for x in rac1p_vals]+SCHL_vals+['CIT_'+x for x in CIT_vals]\
                    +['ESR_'+x for x in ESR_vals]
    
    new_task = BasicProblem(new_features, task._target, task._target_transform,
                task._group, task._group_transform,
                preprocess = add_indicators_pubcov, postprocess = task._postprocess)

    source_X_raw, source_y_raw, _ = new_task.df_to_numpy(source_data)
    
    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(source_X_raw[:,2:])
        source_X_raw[:,2:] = scaler.transform(source_X_raw[:,2:])
        
    return source_X_raw, source_y_raw.astype("int"), new_features

def get_ACSMobility(state, year=2018, need_preprocess=True, root_dir = './datasets/acs'):
    task = ACSMobility
    data_source = ACSDataSource(root_dir=root_dir, survey_year=year, horizon='1-Year', survey='person')
    source_data = data_source.get_data(states=[state], download=False)
    
    source_data = mobility_filter(source_data)
    rac1p_vals = ['white','black','am_ind','alaska','am_alaska','asian','hawaiian','other','two_or_more']
    relp_vals = ['reference', 'husband/wife','biologicalson','adoptedson','stepson','brother','father','grandchild','parentinlaw','soninlaw','other','roomer',
        'housemate','unmarried','foster','nonrelative','institutionalized','noninstitutionalized']
    new_features = ['SEX', 'AGEP', 'SCHL', 'DIS', 'ESP','CIT', 'MIL', 'ANC', 'NATIVITY',  'DEAR','DEYE','DREM',
                    'GCL',  'WKHP', 'JWMNP', 'PINCP', 'married', 'widowed','divorced','separated','never']\
                    +['race_'+x for x in rac1p_vals]+['relp_'+x for x in relp_vals]+COW_vals+['ESR_'+x for x in ESR_vals]
    
    new_task = BasicProblem(new_features, task._target, task._target_transform,
                task._group, task._group_transform,
                preprocess = add_indicators_mobility, postprocess = task._postprocess)

    source_X_raw, source_y_raw, _ = new_task.df_to_numpy(source_data)
    
    if need_preprocess:
        scaler = StandardScaler()
        scaler.fit(source_X_raw[:,2:])
        source_X_raw[:,2:] = scaler.transform(source_X_raw[:,2:])
        
    return source_X_raw, source_y_raw.astype("int"), new_features


def get_data(task, state, need_preprocess, root_dir, year=2018):
    if task == 'income':
        return get_ACSIncome(state, year, need_preprocess, root_dir)
    elif task == 'income_aug':
        return get_ACSIncome_aug(state, year, need_preprocess, root_dir)
    elif task == 'pubcov':
        return get_ACSPubCov(state, year, need_preprocess, root_dir)
    elif task == 'mobility':
        return get_ACSMobility(state, year, need_preprocess, root_dir)
    elif task == 'accident':
        return get_USAccident(state, need_preprocess, root_dir)
    elif task == 'taxi':
        return get_taxi(state, need_preprocess, root_dir)



import numpy as np
import pickle
from acs import ACSDataSource, ACSEmployment, ACSIncome, ACSPublicCoverage, ACSMobility, ACSTravelTime
from folktables import BasicProblem
from folktables import *
import pandas as pd
import random
import seaborn as sns
import pickle





def get_ACSIncome(state, year=2018):
    task = ACSIncome
    data_source = ACSDataSource(root_dir='../', survey_year=year, horizon='1-Year', survey='person')
    source_data = data_source.get_data(states=[state], download=True)
    





if __name__ == "__main__":
    for state in ['CA', 'TX', 'FL', 'NY', 'MT', 'SD']:
        for year in [2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2019, 2021]:
            get_ACSIncome(state, year)
    # get_USAccident('xgb', 'CA')


import concurrent.futures
import os
import pathlib

import requests

from folktables import exceptions


def download_file(url, download_path):
    """Makes a GET request to the specified URL and saves the
    contents of the response to the specified path.

    Parameters
    ----------
    url : str
        URL from where the data will be downloaded.
    download_path : pathlib.Path
        Path to where the downloaded content will be stored.

    Returns
    -------
    download_path : str
        Path to where the content is stored.

    Raises
    ------
    exceptions.FileDownloadError
        This exception is raised if we get an error from our HTTP request
        (i.e., the status code we get from the response is not 200).
    """
    response = requests.get(url)

    response.raise_for_status()

    with open(download_path, 'wb') as handle:
        handle.write(response.content)

    return download_path


def determine_files_to_download(files_resources, download, make_dir=True):
    """Determines which datasets should be downloaded based on whether they
    currently exist in local memory at the specified location.

    Parameters
    ----------
    files_resources : list[FilesResource]
        The FilesResource for the files the user has requested to (down)load.
    download : bool
        Whether the dataset should be downloaded or not.
    make_dir : bool
        Flag to determine whether we should attempt to create a directory
        for every file that has been requested by the user.

    Returns
    -------
    files_to_download : list[FilesResource]
        The datasets that should be downloaded.

    Raises
    ------
    FileNotFoundError
        This error is raised when a file is not found in local memory and
        the user has set the `download` flag to `False`.
    """
    files_to_download = []
    for resource in files_resources:
        if make_dir:
            pathlib.Path(resource.data_dir).mkdir(exist_ok=True,
                                                  parents=True)

        if resource.file_path.is_file():
            continue

        if not download:
            raise FileNotFoundError(
                f'Could not find the file '
                f'{resource.file_name}. Call `get_data` '
                f'with `download=True` to download the '
                f'dataset.'
            )

        files_to_download.append(resource)

    return files_to_download


def download_datasets(files_to_download):
    """Downloads the datasets from their corresponding websites.

    Parameters
    ----------
    files_to_download : list[FilesResource]
        The FilesResources of the files to be downloaded.
    """
    if not files_to_download:
        raise exceptions.NoFilesToDownload(
            'The `files_to_download` list was empty. Make sure that '
            'if there are no files to be downloaded that the function '
            '`download_datasets` is not called.'
        )

    files_names = list(
        map(lambda resource: resource.file_name, files_to_download)
    )
    files_names = ' '.join(files_names)
    print(f'Downloading {len(files_to_download)} file(s): {files_names}')

    if len(files_to_download) > 1:
        # We're currently using the number of threads equivalent to the minimum
        # between the number of CPUs and the number of files to download;
        # however, we can change this to follow a different behavior.
        num_cpus = os.cpu_count()
        if num_cpus is None:
            num_threads = 1
        else:
            num_threads = min(num_cpus, len(files_to_download))

        with concurrent.futures.ThreadPoolExecutor(
                max_workers=num_threads
        ) as executor:
            executor.map(
                lambda resource: download_file(resource.url,
                                               resource.download_path),
                files_to_download
            )

    else:
        resource = files_to_download[0]
        download_file(resource.url, resource.download_path)



from setuptools import setup,find_packages
import os 

here = os.path.abspath(os.path.dirname(__file__))
md_path = os.path.join(here, 'whyshift/README.md')

setup(
    name='whyshift',
    version='0.1.2',    
    description='A package of various specified distribution shift patterns of out-of-distributoin generalization problem on tabular data, and tools for diagnosing model performance are integrated.',
    long_description=open(md_path, encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/namkoong-lab/whyshift',
    author='Jiashuo Liu, Tianyu Wang, Peng Cui, Hongseok Namkoong',
    author_email='liujiashuo77@gmail.com, tw2837@columbia.edu, cuip@tsinghua.edu.cn, namkoong@gsb.columbia.edu',
    packages=find_packages(),
    install_requires=['pandas',
                      'numpy',                     
                      'scikit-learn',
                      'lightgbm','xgboost','fairlearn', 'tqdm', 'torch', 'scipy'
                      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        'Operating System :: POSIX :: Linux',
    ],
    python_requires=">=3",
)


# Built on Folktables and the codebase https://github.com/jpgard/subgroup-robustness-grows-on-trees
__version__ = "0.1.2"
__author__ = 'Jiashuo Liu, Tianyu Wang, Peng Cui, Hongseok Namkoong'
__credits__ = = 'Peng Cui\'s Group and Hongseok Namkoong\' Group'


from .dataset import get_data
from .algorithm import fetch_model
from .disde import degradation_decomp
from .region_analysis import risk_region